import os
import sys
import json
import asyncio
from typing import Optional

import sirius_sdk

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from consts import *


BANK = AGENT2
EMPLOYER = AGENT1
DID_EMPLOYER = 'Th7MpTaRZVRYnPiabds81Y'
VERKEY_EMPLOYER = 'FYmoFw55GeQH7SRFa37dkx1d2dZ3zUF8ckg7wmL7ofN4'
DID_BANK = 'T8MtAB98aCkgNLtNfQx6WG'
VERKEY_BANK = 'FEvX3nsJ8VjW4qQv4Dh9E3NDEx1bUPDtc9vkaaoKVyz1'
CONN_KEY_BANK = 'Hoq65QDfJZ1G4qUHXMPkmXcA51ztaQsVTEHN8rZFpWjv'
CONN_KEY_EMPLOYER = '3huyDF8TmcRr2hRzmyaQH6rqzjFdNy557FBxVXJDpPV4'
DEMO_SALARY = 3000
DEMO_CURRENCY = 'USD'
EMPLOYER_CRED_DEF_ID = 'Th7MpTaRZVRYnPiabds81Y:3:CL:7518:TAG'


def log(message: str):
    print(f'\t{message}')


async def setup_employer_cred_defs(network_name: str = 'test_network'):
    async with sirius_sdk.context(**EMPLOYER):
        dkms = await sirius_sdk.ledger(network_name)  # Test network is prepared for Demo purposes
        schema_id, anon_schema = await sirius_sdk.AnonCreds.issuer_create_schema(
            DID_EMPLOYER, 'demo_salary', '1.0', ['salary', 'currency']
        )
        # Ensure schema exists on DKMS
        schema_ = await dkms.ensure_schema_exists(anon_schema, DID_EMPLOYER)
        # Ensure CredDefs is stored to DKMS
        cred_def_fetched = await dkms.fetch_cred_defs(tag='TAG', schema_id=schema_.id)
        if cred_def_fetched:
            cred_def_ = cred_def_fetched[0]
        else:
            ok, cred_def_ = await dkms.register_cred_def(
                cred_def=sirius_sdk.CredentialDefinition(tag='TAG', schema=schema_),
                submitter_did=DID_EMPLOYER
            )
            assert ok is True
        print('===========================')
        print('cred def ID: ' + cred_def_.id)
        print('===========================')
    return schema_, cred_def_


async def generate_invitations_qr_codes() -> (str, str):
    # BANK
    async with sirius_sdk.context(**BANK):
        try:
            connection_key = await sirius_sdk.Crypto.create_key(seed='0000000000000000SIRIUS_BANK_CONN')
        except sirius_sdk.indy_exceptions.WalletItemAlreadyExists:
            log(f'Bank: conn key {CONN_KEY_BANK} already exists')
        else:
            log(f'Bank: conn key {connection_key} was created')
            assert connection_key == CONN_KEY_BANK
        endpoints = await sirius_sdk.endpoints()
        simple_endpoint = [e for e in endpoints if e.routing_keys == []][0]
        bank_invitation = sirius_sdk.aries_rfc.Invitation(
            label='Sirius Demo bank',
            recipient_keys=[CONN_KEY_BANK],
            endpoint=simple_endpoint.address,
            did=DID_BANK
        )
        log('Bank: invitation')
        log(json.dumps(bank_invitation, indent=2, sort_keys=True))

        # Sirius SDK provide method to generate URL for QR
        # bank_qr_url = await sirius_sdk.generate_qr_code(bank_invitation.invitation_url)
        # print('BANK QR URL: ' + bank_qr_url)

    # EMPLOYER
    async with sirius_sdk.context(**EMPLOYER):
        try:
            connection_key = await sirius_sdk.Crypto.create_key(seed='000000000000SIRIUS_EMPLOYER_CONN')
        except sirius_sdk.indy_exceptions.WalletItemAlreadyExists:
            log(f'Employer: conn key {CONN_KEY_EMPLOYER} already exists')
        else:
            log(f'Employer: conn key {connection_key} was created')
            assert connection_key == CONN_KEY_EMPLOYER
        endpoints = await sirius_sdk.endpoints()
        simple_endpoint = [e for e in endpoints if e.routing_keys == []][0]
        employer_invitation = sirius_sdk.aries_rfc.Invitation(
            label='Sirius Demo employer',
            recipient_keys=[CONN_KEY_EMPLOYER],
            endpoint=simple_endpoint.address,
            did=DID_EMPLOYER
        )
        log('Employer: invitation')
        log(json.dumps(employer_invitation, indent=2, sort_keys=True))

        # Sirius SDK provide method to generate URL for QR
        # employer_qr_url = await sirius_sdk.generate_qr_code(employer_invitation.invitation_url)
        # print('EMPLOYER QR URL: ' + employer_qr_url)

    return 'https://socialsirius.com' + bank_invitation.invitation_url, \
           'https://socialsirius.com' + employer_invitation.invitation_url


async def sirius_bank(network_name: str = 'test_network'):
    async with sirius_sdk.context(**BANK):
        listener = await sirius_sdk.subscribe()
        endpoints = await sirius_sdk.endpoints()
        my_endpoint = [e for e in endpoints if e.routing_keys == []][0]
        dkms = await sirius_sdk.ledger(network_name)
        log('Bank: start to listen events')
        async for event in listener:
            dialog_pairwise: Optional[sirius_sdk.Pairwise] = None
            if isinstance(event.message, sirius_sdk.aries_rfc.ConnRequest):
                # Restore invitation request through invitation.connection_key
                # You may use this snippet to encrypt to invitation cookie values for example
                # to link device to browser Web Page
                if event.recipient_verkey == CONN_KEY_BANK:
                    log(f'Bank: received connection request with connection_key: {CONN_KEY_BANK}')
                    if dialog_pairwise is None:
                        log('Bank: unknown participant. Establish new pairwise P2P')
                    else:
                        log(f'Bank: update pairwise for {dialog_pairwise.their.label}')
                    try:
                        conn_request: sirius_sdk.aries_rfc.ConnRequest = event.message
                        feature_0160 = sirius_sdk.aries_rfc.Inviter(
                            me=sirius_sdk.Pairwise.Me(DID_BANK, VERKEY_BANK),  # Public DID
                            connection_key=event.recipient_verkey,
                            my_endpoint=my_endpoint,
                        )
                        success, dialog_pairwise = await feature_0160.create_connection(conn_request)
                    except Exception as e:
                        log('Bank: exception "%s"' % str(e))
                    else:
                        if success:
                            await sirius_sdk.PairwiseList.ensure_exists(dialog_pairwise)
                            log('Bank: pairwise established successfully')
                            log(json.dumps(dialog_pairwise.metadata, indent=2, sort_keys=True))
                        else:
                            log('Bank: error while establish P2P connection')
                            if feature_0160.problem_report:
                                log('problem report')
                                log(json.dumps(feature_0160.problem_report, indent=2, sort_keys=True))
                else:
                    log('Bank: Unknown connection-key')
            elif isinstance(event.message, sirius_sdk.aries_rfc.Message):
                dialog_pairwise = event.pairwise
            # Digital service
            if dialog_pairwise is not None:
                log('Bank: Digital service')
                person_name = dialog_pairwise.their.label
                service1 = 'Loan request'
                service2 = 'Personal offers'
                ask = sirius_sdk.aries_rfc.Question(
                    valid_responses=[service1, service2],
                    question_text=f'{person_name} welcome to personal cabinet',
                    question_detail='We are glad to make personal offer for you!',
                    locale='en'
                )
                ask.set_ttl(60)  # Set timeout for answer
                success, answer = await sirius_sdk.recipes.ask_and_wait_answer(
                    query=ask,
                    to=dialog_pairwise
                )
                if success:
                    log(f'Bank: providing service: {answer.response}')
                    if answer.response == service1:
                        feature_0037 = sirius_sdk.aries_rfc.Verifier(
                            prover=dialog_pairwise,
                            ledger=dkms,
                        )
                        verify_ok = await feature_0037.verify(
                            comment='Prove your salary credentials',
                            locale='en',
                            proof_request={
                                'nonce': await sirius_sdk.AnonCreds.generate_nonce(),
                                "name": "Proof your salary",
                                "version": "0.1",
                                "requested_attributes": {
                                    'attr1_referent': {
                                        "name": "salary",
                                        "restrictions": {
                                            "issuer_did": DID_EMPLOYER
                                        }
                                    },
                                    'attr2_referent': {
                                        "name": "currency",
                                        "restrictions": {
                                            "issuer_did": DID_EMPLOYER
                                        }
                                    }
                                }
                            },
                            translation=[
                                sirius_sdk.aries_rfc.AttribTranslation('salary', 'Your salary'),
                                sirius_sdk.aries_rfc.AttribTranslation('currency', 'Currency'),
                            ]
                        )
                        if verify_ok:
                            txt_msg = sirius_sdk.aries_rfc.Message(
                                content='Loan approved',
                                locale='en'
                            )
                        else:
                            txt_msg = sirius_sdk.aries_rfc.Message(
                                content='Loan declined',
                                locale='en'
                            )
                        await sirius_sdk.send_to(txt_msg, dialog_pairwise)
                    else:
                        txt_msg = sirius_sdk.aries_rfc.Message(
                            content='Demo personal offer content',
                            locale='en'
                        )
                        await sirius_sdk.send_to(txt_msg, dialog_pairwise)


async def sirius_employer(network_name: str = 'test_network'):
    async with sirius_sdk.context(**EMPLOYER):
        listener = await sirius_sdk.subscribe()
        endpoints = await sirius_sdk.endpoints()
        my_endpoint = [e for e in endpoints if e.routing_keys == []][0]
        dkms = await sirius_sdk.ledger(network_name)
        log('Employer: start to listen events')
        async for event in listener:
            dialog_pairwise: Optional[sirius_sdk.Pairwise] = None
            if isinstance(event.message, sirius_sdk.aries_rfc.ConnRequest):
                # Restore invitation request through invitation.connection_key
                # You may use this snippet to encrypt to invitation cookie values for example
                # to link device to browser Web Page
                if event.recipient_verkey == CONN_KEY_EMPLOYER:
                    log(f'Employer: received connection request with connection_key: {CONN_KEY_BANK}')
                    if dialog_pairwise is None:
                        log('Employer: unknown participant. Establish new pairwise P2P')
                    else:
                        log(f'Employer: update pairwise for {dialog_pairwise.their.label}')
                    try:
                        conn_request: sirius_sdk.aries_rfc.ConnRequest = event.message
                        feature_0160 = sirius_sdk.aries_rfc.Inviter(
                            me=sirius_sdk.Pairwise.Me(DID_EMPLOYER, VERKEY_EMPLOYER),  # Public DID
                            connection_key=event.recipient_verkey,
                            my_endpoint=my_endpoint,
                        )
                        success, dialog_pairwise = await feature_0160.create_connection(conn_request)
                    except Exception as e:
                        log('Employer: exception "%s"' % str(e))
                    else:
                        if success:
                            await sirius_sdk.PairwiseList.ensure_exists(dialog_pairwise)
                            log('Employer: pairwise established successfully')
                            log(json.dumps(dialog_pairwise.metadata, indent=2, sort_keys=True))
                        else:
                            log('Employer: error while establish P2P connection')
                            if feature_0160.problem_report:
                                log('problem report')
                                log(json.dumps(feature_0160.problem_report, indent=2, sort_keys=True))
                else:
                    log('Employer: Unknown connection-key')
            elif isinstance(event.message, sirius_sdk.aries_rfc.Message):
                dialog_pairwise = event.pairwise
            # Digital service
            if dialog_pairwise is not None:
                log('Employer: Digital service')
                person_name = dialog_pairwise.their.label
                service1 = 'Get Salary credentials'
                service2 = 'Request for holiday'
                service3 = 'Go to WebSite'
                ask = sirius_sdk.aries_rfc.Question(
                    valid_responses=[service1, service2, service3],
                    question_text=f'{person_name} welcome!',
                    question_detail='I am your employer Virtual Assistant.',
                    locale='en'
                )
                ask.set_ttl(60)  # Set timeout for answer
                success, answer = await sirius_sdk.recipes.ask_and_wait_answer(
                    query=ask,
                    to=dialog_pairwise
                )
                if success:
                    log(f'Employer: providing service: {answer.response}')
                    if answer.response == service1:
                        log('Employer: starting to issue Salary credential')
                        cred_def = await dkms.load_cred_def(EMPLOYER_CRED_DEF_ID, DID_EMPLOYER)
                        feature_0036 = sirius_sdk.aries_rfc.Issuer(
                            holder=dialog_pairwise,
                        )
                        success = await feature_0036.issue(
                            values={'salary': DEMO_SALARY, 'currency': DEMO_CURRENCY},
                            schema=cred_def.schema,
                            cred_def=cred_def,
                            comment='You may prove your salary in Bank',
                            locale='en',
                            preview=[
                                sirius_sdk.aries_rfc.ProposedAttrib('salary', 'Your salary'),
                                sirius_sdk.aries_rfc.ProposedAttrib('currency', 'Currency')
                            ],
                            cred_id=f'salary-cred-id-for-{dialog_pairwise.their.did}'
                        )
                        if success:
                            log(f'Employer: salary cred successfully finished')
                        else:
                            log('Employer: error while issuing salary credential')
                            if feature_0036.problem_report:
                                log('problem report')
                                log(json.dumps(feature_0036.problem_report, indent=2, sort_keys=True))
                    else:
                        txt_msg = sirius_sdk.aries_rfc.Message(
                            content=f'You selected "{answer.response}" service',
                            locale='en'
                        )
                        await sirius_sdk.send_to(txt_msg, dialog_pairwise)


if __name__ == '__main__':
    qr_bank, qr_employer = asyncio.get_event_loop().run_until_complete(generate_invitations_qr_codes())
    print('invitation URLs')
    print(f'Bank: {qr_bank}')
    print(f'Employer: {qr_employer}')
    print('-------------')
    schema_employer, cred_def_employer = asyncio.get_event_loop().run_until_complete(setup_employer_cred_defs())
    print('*************')
    asyncio.ensure_future(sirius_bank())
    print('#1')
    asyncio.ensure_future(sirius_employer())
    print('#2')
    asyncio.get_event_loop().run_forever()



import datetime
from typing import Optional

import sirius_sdk
from sirius_sdk.hub.coprotocols import CoProtocolThreadedP2P
from sirius_sdk.abstract.p2p import Pairwise
from sirius_sdk.errors.exceptions import SiriusTimeoutIO

from sirius_sdk.agent.aries_rfc.feature_0113_question_answer.messages import Question, Answer


async def ask_and_wait_answer(query: Question, to: Pairwise) -> (bool, Optional[Answer]):
    iso_string = query.expires_time
    if iso_string:
        try:
            expires_at = datetime.datetime.fromisoformat(iso_string)
            expires_at = expires_at.replace(tzinfo=datetime.timezone.utc)
            delta = expires_at.replace(tzinfo=None) - datetime.datetime.utcnow().replace(tzinfo=None)
            ttl = round(delta.total_seconds())
        except Exception as e:
            ttl = None
    else:
        ttl = None

    co = CoProtocolThreadedP2P(
        thid=query.id,
        to=to,
        time_to_live=ttl
    )
    await co.send(query)
    try:
        while True:
            msg, sender_verkey, recipient_verkey = await co.get_one()
            if (sender_verkey == to.their.verkey) and isinstance(msg, Answer):
                return True, msg
    except SiriusTimeoutIO:
        return False, None


async def make_answer(response: str, q: Question, to: Pairwise):
    answer = Answer(
        response=response,
        thread_id=q.id
    )
    answer.set_out_time()
    await sirius_sdk.send_to(answer, to)



import json
import uuid
from typing import Optional

from sirius_sdk.abstract.api import APICrypto
from sirius_sdk.encryption import pack_message as default_pack_message_util, b58_to_bytes


FORWARD = 'https://didcomm.org/routing/1.0/forward'
ENCODING = 'ascii'


async def forward_wired(
        payload: bytes, their_vk: Optional[str], routing_keys: list,
        crypto: APICrypto = None, my_vk: str = None
) -> bytes:
    if my_vk:
        if crypto is None:
            raise RuntimeError('You must pass crypto if my_vk is filled')
    keys_map = {}
    for n in range(len(routing_keys) - 1, 0, -1):  # example: IF routing_keys = ['k1', 'k2', 'k3'] THEN n = [2,1]
        outer_key = routing_keys[n]
        inner_key = routing_keys[n - 1]
        keys_map[outer_key] = inner_key
    keys_map[routing_keys[0]] = their_vk

    for outer_key in routing_keys:
        inner_key = keys_map[outer_key]
        outer_key_bytes = b58_to_bytes(outer_key)
        forwarded = {
            '@id': uuid.uuid4().hex,
            '@type': FORWARD,
            'to': inner_key,
            'msg': json.loads(payload.decode(ENCODING))
        }
        if my_vk:
            payload = await crypto.pack_message(
                json.dumps(forwarded),
                recipient_verkeys=[outer_key],
                sender_verkey=my_vk
            )
        else:
            payload = default_pack_message_util(
                message=json.dumps(forwarded),
                to_verkeys=[outer_key_bytes]
            )
    return payload



from .question_answer import ask_and_wait_answer, make_answer
from .coprotocols import as_called, as_caller
from .confidential_storage import SimpleDataVault, schedule_vaults
from .connections import accept_invitation, InvitationManager

__all__ = [
    "ask_and_wait_answer", "make_answer", "as_called", "as_caller",
    "schedule_vaults", "SimpleDataVault", "accept_invitation", "InvitationManager"
]



from abc import ABC, abstractmethod
from typing import Optional, List, Generator

from sirius_sdk.messaging import Message
from sirius_sdk.errors.exceptions import SiriusConnectionClosed

from .p2p import Pairwise


class Event(Message):

    def __init__(self, pairwise: Pairwise = None, jwe: bytes = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__pairwise = pairwise
        self.__jwe = jwe

    @property
    def pairwise(self) -> Pairwise:
        return self.__pairwise

    @property
    def message(self) -> Optional[Message]:
        if 'message' in self:
            return self['message']
        else:
            return None

    @property
    def recipient_verkey(self) -> Optional[str]:
        return self.get('recipient_verkey', None)

    @property
    def sender_verkey(self) -> Optional[str]:
        return self.get('sender_verkey', None)

    @property
    def forwarded_keys(self) -> List[str]:
        return self.get('forwarded_keys', [])

    @property
    def content_type(self) -> Optional[str]:
        return self.get('content_type', None)

    @property
    def jwe(self) -> Optional[bytes]:
        return self.__jwe

    @property
    def extra(self) -> Optional[dict]:
        return self.get('~extra', None)


class AbstractListener(ABC):

    @abstractmethod
    def is_open(self) -> bool:
        raise NotImplemented

    @abstractmethod
    async def get_one(self, timeout: int = None) -> Event:
        raise NotImplemented

    def __aiter__(self) -> Generator[Event, None, None]:
        if not self.is_open():
            raise SiriusConnectionClosed()
        return self

    async def __anext__(self):
        while True:
            e = await self.get_one()
            return e



""" Message and Module Type related classes and helpers. """
from functools import partial
from operator import is_not
from typing import Union
import re
from semver import VersionInfo, parse

from sirius_sdk.errors.exceptions import SiriusInvalidType

MTURI_RE = re.compile(r'(.*?)([a-z0-9._-]+)/(\d[^/]*)/([a-z0-9._-]+)$')


class Semver(VersionInfo):
    """ Wrapper around the more complete VersionInfo class from semver package.

        This wrapper enables abbreviated versions in message types
        (i.e. 1.0 not 1.0.0).
    """
    SEMVER_RE = re.compile(
        r'^(0|[1-9]\d*)\.(0|[1-9]\d*)(?:\.(0|[1-9]\d*))?$'
    )

    @classmethod
    def from_str(cls, version_str):
        """ Parse version information from a string. """
        matches = Semver.SEMVER_RE.match(version_str)
        if matches:
            args = list(matches.groups())
            if not matches.group(3):
                args.append('0')
            return Semver(*map(int, filter(partial(is_not, None), args)))

        parts = parse(version_str)

        return cls(
            parts['major'],
            parts['minor'],
            parts['patch'],
            parts['prerelease'],
            parts['build']
        )


class Type:
    """ Message and Module type container """
    FORMAT = '{}{}/{}/{}'

    __slots__ = (
        'doc_uri',
        'protocol',
        'version',
        'version_info',
        'name',
        '_normalized',
        '_str'
    )

    def __init__(
            self,
            doc_uri: str,
            protocol: str,
            version: Union[str, Semver],
            name: str):
        if isinstance(version, str):
            try:
                self.version_info = Semver.from_str(version)
            except ValueError as err:
                raise SiriusInvalidType(
                    'Invalid type version {}'.format(version)
                ) from err
            self.version = version
        elif isinstance(version, Semver):
            self.version_info = version
            self.version = str(version)
        else:
            raise SiriusInvalidType(
                '`version` must be instance of str or Semver,'
                ' got {}'.format(type(version).__name__)
            )

        self.doc_uri = doc_uri
        self.protocol = protocol
        self.name = name
        self._str = Type.FORMAT.format(
            self.doc_uri,
            self.protocol,
            self.version,
            self.name
        )
        self._normalized = Type.FORMAT.format(
            self.doc_uri,
            self.protocol,
            self.version_info,
            self.name
        )

    @classmethod
    def from_str(cls, type_str):
        """ Parse type from string. """
        matches = MTURI_RE.match(type_str)
        if not matches:
            raise SiriusInvalidType('Invalid message type')

        return cls(*matches.groups())

    def __str__(self):
        return self._str

    @property
    def normalized(self):
        """ Return the normalized string representation """
        return self._normalized

    def __hash__(self):
        return hash(self._normalized)

    def __eq__(self, other):
        if isinstance(other, Type):
            return self._normalized == other.normalized
        if isinstance(other, str):
            return self._normalized == other
        raise TypeError('Cannot compare Type and {}'.format(type(other)))

    def __ne__(self, other):
        return not self.__eq__(other)



from sirius_sdk.encryption.custom import *
from sirius_sdk.encryption.ed25519 import pack_message, unpack_message
from sirius_sdk.encryption.p2p import P2PConnection


__all__ = [
    "b64_to_bytes", "bytes_to_b64", "b58_to_bytes", "bytes_to_b58", "create_keypair",
    "random_seed", "validate_seed", "pack_message", "unpack_message", "P2PConnection", "sign_message",
    "verify_signed_message", "did_from_verkey"
]



import threading
from typing import Any, List, Optional

from sirius_sdk.abstract.storage import AbstractKeyValueStorage
from sirius_sdk.agent.wallet import RetrieveRecordOptions
from sirius_sdk.agent.wallet.abstract import AbstractNonSecrets


class DefaultNonSecretsStorage(AbstractNonSecrets):

    def __init__(self, storage: AbstractKeyValueStorage):
        self.__storage = storage

    async def add_wallet_record(self, type_: str, id_: str, value: str, tags: dict = None) -> None:
        await self.__storage.select_db(type_)
        meta = await self.__storage.get(id_)
        if meta is not None:
            raise RuntimeError(f'Record with type: {type_} id: {id_} already exists')
        meta = {
            'id': id_,
            'type': type_,
            'value': value,
            'tags': tags or {}
        }
        await self.__storage.set(id_, meta)

    async def update_wallet_record_value(self, type_: str, id_: str, value: str) -> None:
        await self.__storage.select_db(type_)
        meta = await self.__storage.get(id_)
        if meta is None:
            raise RuntimeError(f'Record with type: {type_} id: {id_} does not exists')
        meta['value'] = value
        await self.__storage.set(id_, meta)

    async def update_wallet_record_tags(self, type_: str, id_: str, tags: dict) -> None:
        await self.__storage.select_db(type_)
        meta = await self.__storage.get(id_)
        if meta is None:
            raise RuntimeError(f'Record with type: {type_} id: {id_} does not exists')
        meta['tags'] = tags
        await self.__storage.set(id_, meta)

    async def add_wallet_record_tags(self, type_: str, id_: str, tags: dict) -> None:
        await self.__storage.select_db(type_)
        meta = await self.__storage.get(id_)
        if meta is None:
            raise RuntimeError(f'Record with type: {type_} id: {id_} does not exists')
        stored_tags = meta.get('tags', {})
        meta['tags'] = dict(**tags, **stored_tags)
        await self.__storage.set(id_, meta)

    async def delete_wallet_record_tags(self, type_: str, id_: str, tag_names: List[str]) -> None:
        await self.__storage.select_db(type_)
        meta = await self.__storage.get(id_)
        if meta is None:
            raise RuntimeError(f'Record with type: {type_} id: {id_} does not exists')
        stored_tags = meta.get('tags', {})
        new_tags = {key: val for key, val in stored_tags.items() if key not in tag_names}
        meta['tags'] = new_tags
        await self.__storage.set(id_, meta)

    async def delete_wallet_record(self, type_: str, id_: str) -> None:
        await self.__storage.select_db(type_)
        meta = await self.__storage.get(id_)
        if meta is not None:
            await self.__storage.delete(id_)

    async def get_wallet_record(self, type_: str, id_: str, options: RetrieveRecordOptions) -> Optional[dict]:
        await self.__storage.select_db(type_)
        meta = await self.__storage.get(id_)
        if meta is None:
            raise RuntimeError(f'Record with type: {type_} id: {id_} does not exists')
        ret = self.__build_record(meta, options)
        return ret

    async def wallet_search(self, type_: str, query: dict, options: RetrieveRecordOptions, limit: int = 1) -> (List[dict], int):
        await self.__storage.select_db(type_)
        collection = []
        counter = 0
        d = await self.__storage.items()
        for key, meta in d.items():
            tags = meta.get('tags', {})
            if '$or' in query:
                subs = [{k: v for k, v in tags.items() if k in _or_} for _or_ in query['$or']]
                for sub in subs:
                    success = all([tags[k] == sub[k] for k in sub.keys()])
                    if success:
                        ret = self.__build_record(meta, options)
                        collection.append(ret)
                        counter += 1
                        break
            elif '$in' in query:
                for k, v in tags.items():
                    if v in query['$in']:
                        ret = self.__build_record(meta, options)
                        collection.append(ret)
                        counter += 1
                        break
            else:
                sub = {k: v for k, v in tags.items() if k in query}
                for query_k, query_v in query.items():
                    if type(query_v) is dict and '$in' in query_v:
                        _in_ = query_v['$in']
                        if tags[query_k] in _in_:
                            ret = self.__build_record(meta, options)
                            collection.append(ret)
                            counter += 1
                if sub == query:
                    ret = self.__build_record(meta, options)
                    collection.append(ret)
                    counter += 1
        return collection[:limit], counter

    @staticmethod
    def __build_record(meta: dict, options: RetrieveRecordOptions):
        ret = {'id': meta['id']}
        if options.retrieve_type:
            ret['type'] = meta['type']
        if options.retrieve_tags:
            ret['tags'] = meta['tags']
        if options.retrieve_value:
            ret['value'] = meta['value']
        return ret



from sirius_sdk.messaging.message import Message, register_message_class, restore_message_instance
from sirius_sdk.messaging.type import Type
from sirius_sdk.messaging.validators import validate_common_blocks, check_for_attributes


__all__ = [
    "Message", "Type", "validate_common_blocks", "check_for_attributes",
    "register_message_class", "restore_message_instance"
]



import sys
import asyncio
from typing import Optional, List, Generator

from sirius_sdk.abstract.listener import Event, AbstractListener
from sirius_sdk.agent.connections import AgentEvents
from sirius_sdk.errors.exceptions import SiriusConnectionClosed
from sirius_sdk.messaging import Message, restore_message_instance
from sirius_sdk.agent.pairwise import AbstractPairwiseList
from sirius_sdk.abstract.p2p import Pairwise

PY_35 = sys.version_info >= (3, 5)
PY_352 = sys.version_info >= (3, 5, 2)


class Listener(AbstractListener):

    def __init__(self, source: AgentEvents, pairwise_resolver: AbstractPairwiseList = None):
        self.__source = source
        self.__pairwise_resolver = pairwise_resolver

    def is_open(self) -> bool:
        return self.__source.is_open

    async def get_one(self, timeout: int = None) -> Event:
        event = await self.__source.pull(timeout)
        if 'message' in event:
            ok, message = restore_message_instance(event['message'])
            if ok:
                event['message'] = message
            else:
                event['message'] = Message(event['message'])
        their_verkey = event.get('sender_verkey', None)
        if self.__pairwise_resolver and their_verkey:
            pairwise = await self.__pairwise_resolver.load_for_verkey(their_verkey)
        else:
            pairwise = None
        return Event(pairwise=pairwise, **event)



from dataclasses import dataclass
from typing import Union, Dict, Any, Optional

from sirius_sdk.encryption.p2p import P2PConnection
from sirius_sdk.errors.exceptions import SiriusInitializationError
from sirius_sdk.agent.pairwise import AbstractPairwiseList
from sirius_sdk.abstract.api import APICrypto, APITransport, APIContents, APIDistributedLocks, \
    APICoProtocols, APIRouter, APINetworks
from sirius_sdk.agent.wallet.abstract.did import AbstractDID
from sirius_sdk.agent.wallet.abstract.anoncreds import AbstractAnonCreds
from sirius_sdk.agent.wallet.abstract.cache import AbstractCache
from sirius_sdk.agent.wallet.abstract.non_secrets import AbstractNonSecrets
from sirius_sdk.abstract.storage import AbstractKeyValueStorage
from sirius_sdk.agent.microledgers.abstract import AbstractMicroledgerList


class Config:

    @dataclass
    class Overrides:
        crypto: APICrypto = None
        did: AbstractDID = None
        microledgers: AbstractMicroledgerList = None
        storage: AbstractKeyValueStorage = None
        pairwise_storage: AbstractPairwiseList = None
        non_secrets: AbstractNonSecrets = None
        anoncreds: AbstractAnonCreds = None
        cache: AbstractCache = None
        coprotocols: APICoProtocols = None
        transport: APITransport = None
        contents: APIContents = None
        distr_locks: APIDistributedLocks = None
        router: APIRouter = None
        networks: APINetworks = None

    @dataclass
    class CloudOpts:
        server_uri: str = None
        credentials: Union[str, bytes] = None
        p2p: P2PConnection = None
        io_timeout: int = None

        @property
        def is_filled(self) -> bool:
            return self.server_uri is not None and self.credentials is not None and self.p2p is not None

    @dataclass
    class MediatorOpts:
        uri: str = None
        my_verkey: str = None
        mediator_verkey: str = None

        @property
        def is_filled(self) -> bool:
            return self.uri is not None and self.my_verkey is not None and self.mediator_verkey is not None

    def __init__(self):
        self.__overrides = self.Overrides()
        self.__cloud_opts = self.CloudOpts()
        self.__mediator_opts = self.MediatorOpts()

    @property
    def overrides(self) -> Overrides:
        return self.__overrides

    @property
    def cloud_opts(self) -> CloudOpts:
        return self.__cloud_opts

    @property
    def mediator_opts(self) -> MediatorOpts:
        return self.__mediator_opts

    def setup_cloud(
            self, *args, server_uri: str = None, credentials: Union[bytes, str] = None,
            p2p: Union[P2PConnection, dict] = None, io_timeout: int = None
    ) -> "Config":
        if args:
            if len(args) == 3:
                return self.setup_cloud(server_uri=args[0], credentials=args[1], p2p=args[2])
            else:
                cfg = args[0]
                if type(cfg) is dict:
                    return self.setup_cloud(**cfg)
                else:
                    raise SiriusInitializationError('Expected configuration as dictionary of params')
        # setup cloud agent connection url
        if server_uri is None:
            raise SiriusInitializationError('"server_uri" must be set')
        self.__cloud_opts.server_uri = server_uri
        # setup credentials
        if credentials is None:
            raise SiriusInitializationError('"credentials" must be set')
        elif type(credentials) is str:
            self.__cloud_opts.credentials = credentials.encode()
        elif type(credentials) is bytes:
            self.__cloud_opts.credentials = credentials
        else:
            raise SiriusInitializationError('Unexpected credentials type. Expected str or bytes')
        # setup p2p connection
        if p2p is None:
            raise SiriusInitializationError('"p2p" must be set')
        if type(p2p) is dict:
            their_verkey = p2p.get('their_verkey', None)
            if their_verkey is None:
                raise SiriusInitializationError('p2p does not have "their_verkey" attribute')
            if type(their_verkey) is not str:
                raise SiriusInitializationError('p2p "their_verkey" attribute is not base58 string')
            my_keys = p2p.get('my_keys', None)
            if my_keys is None:
                raise SiriusInitializationError('p2p does not have "my_keys" attribute')
            if type(my_keys) is not list:
                raise SiriusInitializationError('p2p "my_keys" attribute is not base58 keys list')
            if len(my_keys) != 2:
                raise SiriusInitializationError('p2p "my_keys" attribute unexpected structure')
            self.__cloud_opts.p2p = P2PConnection(my_keys=tuple(my_keys), their_verkey=their_verkey)
        elif type(p2p) is P2PConnection:
            self.__cloud_opts.p2p = p2p
        else:
            raise SiriusInitializationError('Unexpected p2p type')
        if io_timeout:
            self.__cloud_opts.io_timeout = io_timeout
        return self

    def setup_mediator(self, uri: str, my_verkey: str, mediator_verkey: str) -> "Config":
        self.__mediator_opts = self.MediatorOpts(uri=uri, my_verkey=my_verkey, mediator_verkey=mediator_verkey)
        return self

    def override(
            self, crypto: APICrypto = None, did: AbstractDID = None,
            microledgers: AbstractMicroledgerList = None, storage: AbstractKeyValueStorage = None,
            pairwise_storage: AbstractPairwiseList = None, non_secrets: AbstractNonSecrets = None,
            anon_cred: AbstractAnonCreds = None, coprotocols: APICoProtocols = None, router: APIRouter = None,
            transport: APITransport = None, contents: APIContents = None,
            distr_locks: APIDistributedLocks = None, networks: APINetworks = None, **kwargs
    ) -> "Config":
        if crypto:
            self.__overrides.crypto = crypto
        if did:
            self.__overrides.did = did
        if microledgers:
            self.__overrides.microledgers = microledgers
        if storage:
            self.__overrides.storage = storage
        if pairwise_storage:
            self.__overrides.pairwise_storage = pairwise_storage
        if non_secrets:
            self.__overrides.non_secrets = non_secrets
        if anon_cred:
            self.__overrides.anon_cred = anon_cred
        if coprotocols:
            self.__overrides.coprotocols = coprotocols
        if transport:
            self.__overrides.transport = transport
        if contents:
            self.__overrides.contents = contents
        if distr_locks:
            self.__overrides.distr_locks = distr_locks
        if router:
            self.__overrides.router = router
        if networks:
            self.__overrides.networks = networks
        return self

    def override_crypto(self, dependency: APICrypto) -> "Config":
        self.__overrides.crypto = dependency
        return self

    def override_did(self, dependency: AbstractDID) -> "Config":
        self.__overrides.did = dependency
        return self

    def override_microledgers(self, dependency: AbstractMicroledgerList) -> "Config":
        self.__overrides.microledgers = dependency
        return self

    def override_storage(self, dependency: AbstractKeyValueStorage) -> "Config":
        self.__overrides.storage = dependency
        return self

    def override_pairwise_storage(self, dependency: AbstractPairwiseList) -> "Config":
        self.__overrides.pairwise_storage = dependency
        return self

    def override_non_secrets(self, dependency: AbstractNonSecrets) -> "Config":
        self.__overrides.non_secrets = dependency
        return self

    def override_anon_cred(self, dependency: AbstractAnonCreds) -> "Config":
        self.__overrides.anon_cred = dependency
        return self

    def override_coprotocols(self, dependency: APICoProtocols) -> "Config":
        self.__overrides.coprotocols = dependency
        return self

    def override_transport(self, dependency: APITransport) -> "Config":
        self.__overrides.transport = dependency
        return self

    def override_contents(self, dependency: APIContents) -> "Config":
        self.__overrides.contents = dependency
        return self

    def override_distr_locks(self, dependency: APIDistributedLocks) -> "Config":
        self.__overrides.distr_locks = dependency
        return self

    def override_router(self, dependency: APIRouter) -> "Config":
        self.__overrides.router = dependency
        return self

    def override_networks(self, dependency: APINetworks) -> "Config":
        self.__overrides.networks = dependency
        return self



from sirius_sdk.agent.aries_rfc.feature_0036_issue_credential.state_machines import Issuer, Holder
from sirius_sdk.agent.aries_rfc.feature_0036_issue_credential.messages import OfferCredentialMessage, \
    RequestCredentialMessage, IssueCredentialMessage, IssueProblemReport, ProposedAttrib, AttribTranslation


__all__ = [
    'Issuer', 'Holder', 'OfferCredentialMessage',
    'RequestCredentialMessage', 'IssueCredentialMessage', 'IssueProblemReport', 'ProposedAttrib', 'AttribTranslation'
]



import asyncio
import logging
from abc import ABC
from typing import Optional, List, Tuple, Dict
from datetime import datetime, timedelta

import sirius_sdk
from sirius_sdk.abstract.bus import AbstractBus
from sirius_sdk.abstract.listener import Event
from sirius_sdk.abstract.batching import RoutingBatch
from sirius_sdk.abstract.p2p import TheirEndpoint, Pairwise
from sirius_sdk.errors.exceptions import *
from sirius_sdk.messaging import Message
from sirius_sdk.messaging.fields import DIDField
from sirius_sdk.errors.exceptions import SiriusContextError, OperationAbortedManually, SiriusTimeoutIO
from sirius_sdk.agent.aries_rfc.base import AriesProtocolMessage
from sirius_sdk.agent.aries_rfc.decorators import PLEASE_ACK_DECORATOR as ARIES_PLEASE_ACK_DECORATOR
from sirius_sdk.agent.aries_rfc.mixins import ThreadMixin, PleaseAckMixin


def _qualify_did(value: str) -> str:
    if ':' not in value:
        return f'did:peer:{value}'
    else:
        return value


class AbstractCoProtocol(ABC):
    """Abstraction application-level protocols in the context of interactions among agent-like things.

        Sirius SDK protocol is high-level abstraction over Sirius transport architecture.
        Approach advantages:
          - developer build smart-contract logic in block-style that is easy to maintain and control
          - human-friendly source code of state machines in procedural style
          - program that is running in separate coroutine: lightweight abstraction to start/kill/state-detection work thread
        See details:
          - https://github.com/hyperledger/aries-rfcs/tree/master/concepts/0003-protocols
    """

    SEC_PER_DAY = 86400
    SEC_PER_HOURS = 3600
    SEC_PER_MIN = 60

    def __init__(self, time_to_live: int = None):
        self.__time_to_live = time_to_live
        self.__is_aborted = False
        self.__die_timestamp = None
        self.__cur_loop: Optional[asyncio.AbstractEventLoop] = None
        self._bus: Optional[AbstractBus] = None
        self._is_running = False
        self._please_ack_ids = []

    def __del__(self):
        if self._is_running and self.__cur_loop and self.__cur_loop.is_running():
            asyncio.ensure_future(self.clean(), loop=self.__cur_loop)
            self.__cur_loop = None

    @property
    def time_to_live(self) -> Optional[int]:
        """Time in seconds protocol will be connected to protocols Events Bus

         if None: protocol is infinite
        """
        return self.__time_to_live

    @property
    def is_alive(self) -> bool:
        """Protocol may running but ttl is out of date
        """
        if not self._is_running:
            return False
        elif self.__die_timestamp:
            return datetime.now() < self.__die_timestamp
        else:
            return True

    @property
    def is_aborted(self) -> bool:
        """Protocol was aborted (forced terminated)"""
        return self.__is_aborted

    async def start(self):
        if self.__time_to_live:
            self.__die_timestamp = datetime.now() + timedelta(seconds=self.__time_to_live)
        else:
            self.__die_timestamp = None
        self.__cur_loop = asyncio.get_event_loop()
        self._bus = await sirius_sdk.spawn_coprotocol()
        self._is_running = True

    async def stop(self):
        self._is_running = False
        self.__cur_loop = None
        self._bus = None
        self._please_ack_ids.clear()

    async def abort(self):
        """Useful to Abort running co-protocol outside of current loop"""
        if not self.__is_aborted:
            self.__is_aborted = True
            await self.__abort()

    async def clean(self):
        if self._is_running:
            await self.stop()
            self._is_running = False

    def _get_io_timeout(self):
        if self.__die_timestamp:
            now = datetime.now()
            if now < self.__die_timestamp:
                delta = self.__die_timestamp - now
                return delta.days * self.SEC_PER_DAY + delta.seconds
            else:
                return 0
        else:
            return None

    async def _setup_context(self, message: Message):
        if isinstance(message, PleaseAckMixin):
            message.please_ack = True
        ack_message_id = self._extract_ack_id(message)
        if ack_message_id and ack_message_id not in self._please_ack_ids:
            await self._bus.subscribe(ack_message_id)
            self._please_ack_ids.append(ack_message_id)

    async def _cleanup_context(self, message: Message = None):
        if message:
            ack_message_id = self._extract_ack_id(message)
            if ack_message_id:
                await self._bus.unsubscribe(ack_message_id)
                self._please_ack_ids = [i for i in self._please_ack_ids if i != ack_message_id]
            thread = ThreadMixin.get_thread(message)
            if thread:
                if thread.thid in self._please_ack_ids:
                    await self._bus.unsubscribe(thread.thid)
                    self._please_ack_ids = [i for i in self._please_ack_ids if i != thread.thid]
        else:
            await self._bus.unsubscribe_ext(self._please_ack_ids)
            self._please_ack_ids.clear()

    @staticmethod
    def _extract_ack_id(message: Message) -> Optional[str]:
        if isinstance(message, AriesProtocolMessage):
            if message.please_ack is True:
                return message.ack_message_id
        else:
            # If message has not-registered type
            if ARIES_PLEASE_ACK_DECORATOR in message:
                aries_msg = AriesProtocolMessage(**message)
                return aries_msg.ack_message_id
        return None

    async def __abort(self):
        if self._bus:
            await self._bus.abort()


class AbstractP2PCoProtocol(AbstractCoProtocol):

    def __init__(self, protocols: List[str] = None, time_to_live: int = None):
        super().__init__(time_to_live)
        self.__their_vk: Optional[str] = None
        self.__endpoint: Optional[str] = None
        self.__my_vk: Optional[str] = None
        self.__routing_keys: Optional[List[str]] = None
        self.__protocols = protocols
        self.__is_setup = False
        self.__ack_thread_id = None
        self._binding_ids = []

    @property
    def binding_ids(self) -> List[str]:
        return self._binding_ids

    async def start(self):
        if not self.__is_setup:
            raise SiriusPendingOperation('You must Setup protocol instance at first')
        await super().start()
        await self._subscribe_to_events()
        # Remove duplicates
        self._binding_ids = list(set(self._binding_ids))

    async def stop(self):
        if self._binding_ids or self._please_ack_ids:
            await self._bus.unsubscribe_ext(self._binding_ids + self._please_ack_ids)
            self._binding_ids.clear()
            self._please_ack_ids.clear()
        await super().stop()

    async def send(self, message: Message):
        if not self.__is_setup:
            raise SiriusPendingOperation('You must Setup protocol instance at first')
        if not self._is_running:
            await self.start()
        await self._before(message, include_please_ack=False)
        await self._setup_context(message)
        await sirius_sdk.send(
            message=message, their_vk=self.__their_vk, endpoint=self.__endpoint,
            my_vk=self.__my_vk, routing_keys=self.__routing_keys
        )

    async def get_one(self) -> (Optional[Message], str, Optional[str]):
        """Accumulate event from participant

        return: message, sender_verkey, recipient_verkey
        """
        if not self._is_running:
            await self.start()
        expected_binding_ids = self._binding_ids + self._please_ack_ids
        while True:
            # re-calc timeout in loop until event income
            timeout = self._get_io_timeout()
            if (timeout is not None) and (timeout <= 0):
                raise SiriusTimeoutIO
            # wait
            try:
                event = await self._bus.get_message(timeout=timeout)
            except OperationAbortedManually:
                await self.clean()
                raise
            # process event
            await self._after(event.message)
            if event.thread_id in expected_binding_ids:
                return event.message, event.sender_verkey, event.recipient_verkey
            else:
                # co-protocols are competitors, so
                # they operate in displace multitasking mode
                logging.warning(
                    f'Expected binding_id: "{expected_binding_ids}" actually "{event.thread_id}" was received'
                )

    async def switch(self, message: Message) -> (bool, Message):
        """Send Message to other-side of protocol and wait for response

        :param message: Protocol request
        :return: (success, Response)
        """
        if not self.__is_setup:
            raise SiriusPendingOperation('You must Setup protocol instance at first')
        if not self._is_running:
            await self.start()
        try:
            await self._before(message, include_please_ack=True)
            await self._setup_context(message)
            try:
                await self.send(message)
                message, sender_verkey, recipient_verkey = await self.get_one()
                await self._after(message)
                return True, message
            finally:
                await self._cleanup_context(message)
        except SiriusTimeoutIO:
            return False, None

    async def _subscribe_to_events(self):
        ok, binding_ids = await self._bus.subscribe_ext(
            sender_vk=[self.__their_vk], recipient_vk=[self.__my_vk], protocols=self.__protocols
        )
        if ok:
            self._binding_ids.extend(binding_ids)
            self._binding_ids = list(set(self._binding_ids))
        else:
            raise SiriusRPCError('Error with subscribe to protocol events')

    def _setup(self, their_verkey: str, endpoint: str, my_verkey: str = None, routing_keys: List[str] = None):
        """Should be called in Descendant"""
        self.__their_vk = their_verkey
        self.__my_vk = my_verkey
        self.__endpoint = endpoint
        self.__routing_keys = routing_keys or []
        self.__is_setup = True

    async def _before(self, message: Message, include_please_ack: bool = True):
        if isinstance(message, AriesProtocolMessage):
            if include_please_ack and message.please_ack is False:
                message.please_ack = True
            if self.__ack_thread_id and not message.thread_id:  # Don't rewrite earlier set values
                thread = ThreadMixin.get_thread(message) or ThreadMixin.Thread()
                thread.thid = self.__ack_thread_id
                ThreadMixin.set_thread(message, thread)
                self.__ack_thread_id = None

    async def _after(self, message: Message):
        ack_id = self._extract_ack_id(message)
        if ack_id:
            self.__ack_thread_id = ack_id
        else:
            self.__ack_thread_id = None


class CoProtocolP2PAnon(AbstractP2PCoProtocol):

    def __init__(self, my_verkey: str, endpoint: TheirEndpoint, protocols: List[str], time_to_live: int = None):
        if not protocols:
            raise SiriusContextError('You must set protocols list. It is empty for now!')
        super().__init__(protocols=protocols, time_to_live=time_to_live)
        self._setup(
            their_verkey=endpoint.verkey, endpoint=endpoint.address,
            my_verkey=my_verkey, routing_keys=endpoint.routing_keys
        )


class CoProtocolP2P(AbstractP2PCoProtocol):

    def __init__(self, pairwise: Pairwise, protocols: List[str], time_to_live: int = None):
        if not protocols:
            raise SiriusContextError('You must set protocols list. It is empty for now!')
        super().__init__(protocols=protocols, time_to_live=time_to_live)
        self._setup(their_verkey=pairwise.their.verkey, endpoint=pairwise.their.address, my_verkey=pairwise.me.verkey)


class CoProtocolThreadedP2P(AbstractP2PCoProtocol):

    def __init__(self, thid: str, to: Pairwise, pthid: str = None, time_to_live: int = None):
        super().__init__(time_to_live=time_to_live)
        self.__thid = thid
        self.__pthid = pthid
        self.__to = to
        self.__sender_order = 0
        self.__received_orders = {}
        self.__last_received_msg_id = {}
        recipient = _qualify_did(to.their.did)
        self.__received_orders[recipient] = 0
        self._setup(their_verkey=to.their.verkey, endpoint=to.their.address, my_verkey=to.me.verkey)

    @property
    def thid(self) -> str:
        return self.__thid

    async def _subscribe_to_events(self):
        ok = await self._bus.subscribe(thid=self.__thid)
        if ok:
            self._binding_ids.append(self.__thid)
            self._binding_ids = list(set(self._binding_ids))
        else:
            raise SiriusRPCError('Error with subscribe to protocol events')

    async def _before(self, message: Message, include_please_ack: bool = True):
        await super()._before(message, include_please_ack)
        thread = ThreadMixin.get_thread(message)
        if thread is None:  # Don't rewrite externally set thread values
            thread = ThreadMixin.Thread(
                thid=self.__thid, pthid=self.__pthid,
                sender_order=self.__sender_order, received_orders=dict(**self.__received_orders)
            )
            self.__sender_order += 1
            ThreadMixin.set_thread(message, thread)

    async def _after(self, message: Message):
        await super()._after(message)
        if self.__to is not None:
            recipient = _qualify_did(self.__to.their.did)
            err = DIDField().validate(recipient)
            if err is None:
                last_msg_id = self.__last_received_msg_id.get(recipient, None)
                if last_msg_id != message.id:
                    if recipient not in self.__received_orders:
                        order = 1
                    else:
                        order = self.__received_orders[recipient] + 1
                    self.__received_orders[recipient] = order
                    self.__last_received_msg_id[recipient] = message.id


class CoProtocolThreadedTheirs(AbstractCoProtocol):

    def __init__(self, thid: str, theirs: List[Pairwise], pthid: str = None, time_to_live: int = None):
        if len(theirs) < 1:
            raise SiriusContextError('theirs is empty')
        super().__init__(time_to_live=time_to_live)
        self.__thid = thid
        self.__pthid = pthid
        self.__theirs = theirs
        self.__dids = [their.their.did for their in theirs]
        self.__received_orders = {}
        for to in theirs:
            recipient = _qualify_did(to.their.did)
            self.__received_orders[recipient] = 0
        self.__last_received_msg_id = {}
        self.__sender_order = 0

    @property
    def theirs(self) -> List[Pairwise]:
        return self.__theirs

    async def start(self):
        await super().start()
        ok = await self._bus.subscribe(thid=self.__thid)
        if not ok:
            raise SiriusRPCError('Error with subscribe to protocol events')

    async def stop(self):
        binding_ids = [self.__thid] + self._please_ack_ids
        await self._bus.unsubscribe_ext(binding_ids)
        self._please_ack_ids.clear()
        await super().stop()

    async def send(self, message: Message) -> Dict[Pairwise, Tuple[bool, str]]:
        """Send message to given participants

        return: List[( str: participant-id, bool: message was successfully sent, str: endpoint response body )]
        """

        if not self._is_running:
            await self.start()

        batches = [
            RoutingBatch(p.their.verkey, p.their.endpoint, p.me.verkey, p.their.routing_keys)
            for p in self.__theirs
        ]

        await self._before(message)
        await self._setup_context(message)
        responses = await sirius_sdk.send_batched(message, batches)
        results = {}
        for p2p, response in zip(self.__theirs, responses):
            success, body = response
            results[p2p] = (success, body)
        return results

    async def get_one(self) -> Tuple[Optional[Pairwise], Optional[Message]]:
        """Read event from any of participants at given timeout

        return: (Pairwise: participant-id, Message: message from given participant)
        """
        if not self._is_running:
            await self.start()
        expected_thread_ids = [self.__thid] + self._please_ack_ids
        while True:
            # re-calc timeout in loop until event income
            timeout = self._get_io_timeout()
            if (timeout is not None) and (timeout <= 0):
                raise SiriusTimeoutIO
            # wait
            try:
                event = await self._bus.get_message(timeout=timeout)
            except SiriusTimeoutIO:
                return None, None
            except OperationAbortedManually:
                await self.clean()
                raise
            # process event
            if event.thread_id in expected_thread_ids:
                p2p = self.__load_p2p_from_verkey(event.sender_verkey)
                return p2p, event.message
            else:
                # co-protocols are competitors, so
                # they operate in displace multitasking mode
                logging.warning(
                    f'Expected thread_id: "{expected_thread_ids}" actually "{event.thread_id}" was received'
                )

    async def switch(self, message: Message) -> Dict[Pairwise, Tuple[bool, Optional[Message]]]:
        """Switch state while participants at given timeout give responses

        return: {
            Pairwise: participant,
            (
              bool: message was successfully sent to participant,
              Message: response message from participant or Null if request message was not successfully sent
            )
        }
        """
        statuses = await self.send(message)
        # fill errors to result just now
        results = {p2p: (False, None) for p2p, stat in statuses.items() if stat[0] is True}
        # then work with success participants only
        success_theirs = {p2p: (False, None) for p2p, stat in statuses.items() if stat[0] is True}
        accum = 0
        rcv_messages = []
        try:
            while accum < len(success_theirs):
                p2p, message = await self.get_one()
                if p2p is None:
                    break
                await self._after(message, p2p)
                if p2p and p2p.their.did in self.__dids:
                    success_theirs[p2p] = (True, message)
                    accum += 1
                rcv_messages.append(message)
            results.update(success_theirs)
            return results
        finally:
            await self._cleanup_contexts(rcv_messages)

    async def _before(self, message: Message):
        thread = ThreadMixin.get_thread(message)
        if thread is None:  # Don't rewrite externally set thread values
            thread = ThreadMixin.Thread(
                thid=self.__thid, pthid=self.__pthid,
                sender_order=self.__sender_order, received_orders=dict(**self.__received_orders)
            )
            ThreadMixin.set_thread(message, thread)

    async def _after(self, message: Message, p2p: Pairwise):
        recipient = _qualify_did(p2p.their.did)
        err = DIDField().validate(recipient)
        if err is None:
            last_msg_id = self.__last_received_msg_id.get(recipient, None)
            if last_msg_id != message.id:
                if recipient not in self.__received_orders:
                    order = 1
                else:
                    order = self.__received_orders[recipient] + 1
                self.__received_orders[recipient] = order
                self.__last_received_msg_id[recipient] = message.id

    async def _cleanup_contexts(self, messages: List[Message]):
        for msg in messages:
            await self._cleanup_context(msg)

    def __load_p2p_from_verkey(self, verkey: str) -> Optional[Pairwise]:
        for p2p in self.__theirs:
            if p2p.their.verkey == verkey:
                return p2p
        return None


def prepare_response(request: Message, response: Message):
    thread_id = None
    parent_thread_id = None
    thread = ThreadMixin.get_thread(request)
    if thread and thread.thid:
        thread_id = thread.thid
    ack_id = PleaseAckMixin.get_ack_message_id(request)
    if ack_id:
        parent_thread_id = thread_id
        thread_id = ack_id
    if thread_id:
        ThreadMixin.set_thread(
            response,
            value=ThreadMixin.Thread(thid=thread_id, pthid=parent_thread_id)
        )


async def open_communication(event: Event, time_to_live: int = None) -> Optional[AbstractP2PCoProtocol]:
    if event.pairwise is not None and event.message is not None:
        thread_id = None
        parent_thread_id = None
        thread = ThreadMixin.get_thread(event.message)
        if thread and thread.thid:
            thread_id = thread.thid
        ack_id = PleaseAckMixin.get_ack_message_id(event.message)
        if ack_id:
            parent_thread_id = thread_id
            thread_id = ack_id
        if thread_id:
            comm = CoProtocolThreadedP2P(
                thid=thread_id,
                to=event.pairwise,
                pthid=parent_thread_id,
                time_to_live=time_to_live
            )
        else:
            comm = CoProtocolP2P(
                pairwise=event.pairwise,
                protocols=[event.message.protocol],
                time_to_live=time_to_live
            )
        return comm
    else:
        return None



import uuid

import pytest

from sirius_sdk.errors.exceptions import BaseSiriusException
from sirius_sdk import Agent, Pairwise
from sirius_sdk.messaging import restore_message_instance
from sirius_sdk.agent.aries_rfc.feature_0753_bus import *

from tests.helpers import run_coroutines, IndyAgent


@pytest.mark.asyncio
async def test_messages():
    cast1 = BusOperation.Cast(thid='some-thread-id')
    assert cast1.validate() is True
    cast2 = BusOperation.Cast(recipient_vk='VK1', sender_vk='VK2', protocols=['some-protocol'])
    assert cast2.validate() is True
    err_cast = BusOperation.Cast(recipient_vk='VK1', sender_vk='VK2')
    assert err_cast.validate() is False

    op_subscribe1 = BusSubscribeRequest(cast1)
    assert op_subscribe1.cast.thid == 'some-thread-id'
    assert op_subscribe1.cast.sender_vk is None and op_subscribe1.cast.recipient_vk is None
    assert op_subscribe1.thread_id is None

    op_subscribe2 = BusSubscribeRequest(cast2)
    assert op_subscribe2.cast.thid is None
    assert op_subscribe2.cast.sender_vk == 'VK2'
    assert op_subscribe2.cast.recipient_vk == 'VK1'

    bind = BusBindResponse(thread_id='some-bind-id')
    assert bind.thread_id == 'some-bind-id'
    assert bind.parent_thread_id is None

    ok, msg = restore_message_instance(
        {
            '@type': 'https://didcomm.org/bus/1.0/bind',
            '~thread': {'thid': 'some-binding-id2'}
        }
    )
    assert ok is True
    assert isinstance(msg, BusBindResponse)
    assert msg.thread_id == 'some-binding-id2'
    assert msg.parent_thread_id is None

    op_subscribe_client_id = BusSubscribeRequest(cast1, parent_thread_id='client-id')
    assert op_subscribe_client_id.parent_thread_id == 'client-id'
    op_bind_client_id = BusBindResponse(thread_id='some-bind-id', parent_thread_id='client-id')
    assert op_bind_client_id.parent_thread_id == 'client-id'
    op_unsubscr_client_id = BusUnsubscribeRequest(client_id='client-id', aborted=True)
    assert op_unsubscr_client_id.client_id == 'client-id'
    assert op_unsubscr_client_id.aborted is True



import io
import math
import os
import os.path
import struct
from typing import Optional, List, Coroutine
from urllib.parse import urlparse

import aiofiles

from sirius_sdk.agent.aries_rfc.feature_0750_storage import AbstractReadOnlyStream, StreamDecryption, \
    StreamInitializationError, ConfidentialStorageEncType, EncryptionError, StreamSeekableError, StreamEOF, \
    StreamFormatError, AbstractWriteOnlyStream, StreamEncryption, BaseStreamEncryption
from sirius_sdk.agent.aries_rfc.feature_0750_storage.components import ConfidentialStorageRawByteStorage, ConfidentialStorageAuthProvider
from sirius_sdk.errors.exceptions import SiriusInitializationError
from sirius_sdk.hub import _current_hub


class FileSystemReadOnlyStream(AbstractReadOnlyStream):

    def __init__(self, path: str, chunks_num: int, enc: Optional[StreamDecryption] = None):
        if chunks_num < 0:
            raise StreamInitializationError('Chunks Num must be greater or equal to 0')
        super().__init__(path, chunks_num, enc)
        self.__fd = None
        self.__size = 0
        self.__chunk_size = 0
        self.__chunk_offsets = []
        self.__on_closed: Optional[Coroutine] = None
        if enc:
            if enc.type != ConfidentialStorageEncType.UNKNOWN:
                if enc.cek is None:
                    try:
                        hub = _current_hub()
                        crypto_manager_exists = hub.get_crypto() is not None
                        if crypto_manager_exists is False:
                            raise EncryptionError(
                                'Crypto manager is not configured'
                            )
                    except SiriusInitializationError:
                        raise EncryptionError(
                            'You should initialize SDK or Call setup() to manually pass keys for decoder'
                        )

    @property
    def on_closed(self) -> Optional[Coroutine]:
        return self.__on_closed

    @on_closed.setter
    def on_closed(self, cb: Coroutine):
        self.__on_closed = cb

    async def open(self):
        if self.__fd:
            return
        self.__fd = await aiofiles.open(self.path, 'rb')
        self.__size = await self.__fd.seek(0, io.SEEK_END)
        await self.__fd.seek(0, io.SEEK_SET)
        if self.enc:
            await self.__load_enc_chunks()
            if self.__chunk_offsets:
                self.__chunk_size = self.__chunk_offsets[0][0]
            else:
                self.__chunk_size = 0
        else:
            self.__chunk_size = math.ceil(self.__size / self.chunks_num)
        self._seekable = True
        self._is_open = True

    async def close(self):
        if self.__fd:
            await self.__fd.close()
            self.__size = 0
            self.__fd = None
            self._seekable = None
            self._is_open = False
            self.__chunk_offsets.clear()
            if self.__on_closed:
                await self.__on_closed

    async def seek_to_chunk(self, no: int) -> int:
        self.__assert_is_open()
        if not self._seekable:
            raise StreamSeekableError('Stream is not seekable')
        pos = no * self.__chunk_size
        if pos > self.__size:
            raise StreamEOF('EOF')
        else:
            new_pos = await self.__fd.seek(pos, io.SEEK_SET)
            self._current_chunk = math.trunc(new_pos / self.__chunk_size)
            return self._current_chunk

    async def read_chunk(self, no: int = None) -> (int, bytes):
        """Read next chunk

        :param no: int (optional) chunk offset, None if reading from current offset
        :raises StreamEOF if end of stream

        :return (new-chunk-offset, data)
        """
        self.__assert_is_open()
        if no is not None:
            await self.seek_to_chunk(no)
        if self._current_chunk >= self.chunks_num:
            raise StreamEOF('EOF')
        file_pos = self._current_chunk * self.__chunk_size
        if self.enc:
            if self._current_chunk >= self.chunks_num:
                raise StreamEOF('EOF')
            info = self.__chunk_offsets[self._current_chunk]
            sz_to_read = info[0]
            seek_to = info[1]
            await self.__fd.seek(seek_to)
            encrypted = await self.__fd.read(sz_to_read)
            if len(encrypted) != sz_to_read:
                raise StreamFormatError('Unexpected encoded file structure')
            # Decode bytes stream
            decrypted = await self.decrypt(encrypted)
            file_pos += len(encrypted)
            self._current_chunk += 1
            return self._current_chunk, decrypted
        else:
            raw = await self.__fd.read(self.__chunk_size)
            self._current_chunk += 1
            return self._current_chunk, raw

    async def eof(self) -> bool:
        self.__assert_is_open()
        if self._current_chunk < self.chunks_num:
            return False
        else:
            return True

    async def __load_enc_chunks(self):
        file_pos = 0
        self.__chunk_offsets.clear()
        while file_pos < self.__size:
            b = await self.__fd.read(4)
            file_pos += len(b)
            if len(b) != 4:
                raise StreamFormatError('Unexpected encoded file structure')
            sz = struct.unpack("i", b)[0]
            self.__chunk_offsets.append([sz, file_pos])
            file_pos = await self.__fd.seek(sz, io.SEEK_CUR)

    def __assert_is_open(self):
        if not self.__fd:
            raise StreamInitializationError('FileStream is not Opened!')


class FileSystemWriteOnlyStream(AbstractWriteOnlyStream):

    DEF_CHUNK_SIZE = 1024  # 1KB

    def __init__(self, path: str, chunk_size: int = DEF_CHUNK_SIZE, enc: Optional[StreamEncryption] = None):
        super().__init__(path, chunk_size, enc)
        self.__cek = None
        self.__file_size = 0
        self.__file_pos = 0
        self.__chunk_offsets = []
        self.__on_closed: Optional[Coroutine] = None
        if enc:
            if enc.type != ConfidentialStorageEncType.UNKNOWN:
                if enc.type != ConfidentialStorageEncType.X25519KeyAgreementKey2019:
                    raise EncryptionError(f'Unsupported key agreement "{enc.type}"')
                if not enc.recipients:
                    raise EncryptionError(f'Recipients data missed, call Setup first!')
        self.__fd = None

    @property
    def on_closed(self) -> Optional[Coroutine]:
        return self.__on_closed

    @on_closed.setter
    def on_closed(self, cb: Coroutine):
        self.__on_closed = cb

    async def create(self, truncate: bool = False):
        async with aiofiles.open(self.path, 'w+b') as fd:
            if truncate:
                await fd.truncate(0)

    async def open(self):
        self.__fd = await aiofiles.open(self.path, 'a+b', buffering=0)
        file_is_seekable = await self.__fd.seekable()
        self._seekable = file_is_seekable
        self.__file_pos = await self.__fd.seek(0, io.SEEK_END)
        self.__file_size = self.__file_pos
        if self.enc:
            await self.__fd.seek(0, io.SEEK_SET)
            await self.__load_enc_chunks()
            await self.__fd.seek(0, io.SEEK_END)
            self._chunks_num = len(self.__chunk_offsets)
            self._current_chunk = self._chunks_num
        else:
            self._current_chunk = math.trunc(self.__file_pos / self._chunk_size)
            self._chunks_num = self._current_chunk
        self._is_open = True

    async def close(self):
        if self.__fd:
            await self.__fd.flush()
            await self.__fd.close()
            self.__fd = None
            self._seekable = None
            self.__file_size = 0
            self.__file_pos = 0
            self._is_open = False
            self.__chunk_offsets.clear()
            if self.__on_closed:
                await self.__on_closed
                self.__on_closed = None

    async def seek_to_chunk(self, no: int) -> int:
        self.__assert_is_open()
        if no == self._current_chunk:
            return no
        elif no == self._chunks_num:
            self._current_chunk = self._chunks_num
            return self._current_chunk
        elif no > self._chunks_num:
            raise StreamEOF('EOF')
        if self.__chunk_offsets:
            try:
                sz, file_pos = self.__chunk_offsets[no]
            except IndexError:
                raise StreamEOF('Chunk No out of range')
        else:
            file_pos = no * self._chunk_size
        if file_pos > self.__file_size:
            raise StreamEOF('EOF')
        else:
            file_pos = await self.__fd.seek(file_pos, io.SEEK_SET)
            self._current_chunk = math.trunc(file_pos / self._chunk_size)
            self.__file_pos = file_pos
            return self._current_chunk

    async def write_chunk(self, chunk: bytes, no: int = None) -> (int, int):
        self.__assert_is_open()
        if no is not None:
            await self.seek_to_chunk(no)
        if self.enc:
            # encode
            encoded = await self.encrypt(chunk)
            # Write Chunk Header with actual encoded bytes size
            sz = len(encoded)
            offset1 = await self.__fd.write(struct.pack("i", sz))
            offset2 = await self.__fd.write(encoded)
            offset = offset1 + offset2
            if no is not None and no < len(self.__chunk_offsets):
                self.__chunk_offsets[no] = (sz, self.__file_pos)
            else:
                self.__chunk_offsets.append((sz, self.__file_pos))
        else:
            offset = await self.__fd.write(chunk)
        self.__file_pos += offset
        if self.__file_pos >= self.__file_size:
            self.__file_size += self.__file_pos - self.__file_size
            self._chunks_num += 1
        self._current_chunk += 1
        return self._current_chunk, len(chunk)

    async def truncate(self, no: int = 0):
        self.__assert_is_open()
        if no == 0:
            await self.__fd.truncate(0)
            await self.__fd.flush()
            self._current_chunk = 0
            self._chunks_num = 0
        else:
            if self.__chunk_offsets:
                max_chunk_offset = len(self.__chunk_offsets)
                if no >= max_chunk_offset:
                    return
                else:
                    sz, file_pos = self.__chunk_offsets[no]
                    await self.__fd.truncate(file_pos)
                    await self.__fd.flush()
                    self.__chunk_offsets = self.__chunk_offsets[:no]
                    self._chunks_num = len(self.__chunk_offsets)
                    if self._current_chunk > no:
                        await self.seek_to_chunk(no)
            else:
                file_pos = no * self._chunk_size
                if file_pos >= self.__file_size:
                    return self._chunks_num
                else:
                    await self.__fd.truncate(file_pos)
                    await self.__fd.flush()
                    self._chunks_num = no
                    if self._current_chunk > no:
                        await self.seek_to_chunk(no)

    async def __load_enc_chunks(self):
        file_pos = 0
        self.__chunk_offsets.clear()
        while file_pos < self.__file_size:
            b = await self.__fd.read(4)
            file_pos += len(b)
            if len(b) != 4:
                raise StreamFormatError('Unexpected encoded file structure')
            sz = struct.unpack("i", b)[0]
            self.__chunk_offsets.append([sz, file_pos])
            file_pos = await self.__fd.seek(sz, io.SEEK_CUR)

    def __assert_is_open(self):
        if not self.__fd:
            raise StreamInitializationError('FileStream is not Opened')


class FileSystemRawByteStorage(ConfidentialStorageRawByteStorage):

    def __init__(self, encryption: BaseStreamEncryption = None):
        super().__init__(encryption)
        self.__mount_dir: Optional[str] = None

    async def mount(self, path: str):
        self.__mount_dir = path

    async def create(self, uri: str):
        path = self.__uri_to_path(uri)
        if os.path.isfile(path):
            raise StreamInitializationError(f'Stream with URI: {uri} already exists!')
        else:
            dirname = os.path.dirname(path)
            if not os.path.isdir(dirname):
                # Ensure nested dirs exists
                try:
                    os.makedirs(dirname, exist_ok=True)
                except OSError as e:
                    raise StreamInitializationError(*e.args)
            stream = FileSystemWriteOnlyStream(path)
            await stream.create()

    async def remove(self, uri: str):
        path = self.__uri_to_path(uri)
        if os.path.isfile(path):
            os.remove(path)

    async def readable(self, uri: str, chunks_num: int) -> AbstractReadOnlyStream:
        path = self.__uri_to_path(uri)
        if not os.path.isfile(path):
            raise StreamInitializationError(f'Stream with URI: {uri} doe not exists!')
        return FileSystemReadOnlyStream(path=path, chunks_num=chunks_num, enc=self.encryption)

    async def writeable(self, uri: str) -> AbstractWriteOnlyStream:
        path = self.__uri_to_path(uri)
        if not os.path.isfile(path):
            raise StreamInitializationError(f'Stream with URI: {uri} doe not exists!')
        return FileSystemWriteOnlyStream(path=path, enc=self.encryption)

    async def exists(self, uri: str) -> bool:
        path = self.__uri_to_path(uri)
        return os.path.isfile(path)

    def __uri_to_path(self, uri: str) -> str:
        p = urlparse(uri)
        path = os.path.join(p.netloc, p.path)
        while path.startswith('/'):
            path = path[1:]
        if self.__mount_dir is not None:
            path = os.path.realpath(os.path.join(self.__mount_dir, path))
        return path



import asyncio
import os
import uuid
from os.path import isfile
from urllib.parse import urlparse

import pytest

import sirius_sdk
from sirius_sdk.messaging import Message
from sirius_sdk.hub.defaults.default_apis import APIDefault

from tests.helpers import ServerTestSuite, run_coroutines


@pytest.mark.asyncio
async def test_qr_codes():
    api = APIDefault()
    content = 'some-content/'*10
    url = await api.generate_qr_code(content)
    p = urlparse(url)
    paths = [p.path] + [p.path[1:]]
    assert any([isfile(path) for path in paths])
    # Clean
    for path in paths:
        if isfile(path):
            os.remove(path)


@pytest.mark.asyncio
async def test_send(test_suite: ServerTestSuite):
    sender_params = test_suite.get_agent_params('agent1')
    receiver_params = test_suite.get_agent_params('agent2')
    sender_verkey = list(sender_params['entities'].items())[0][1]['verkey']
    receiver_verkey = list(receiver_params['entities'].items())[0][1]['verkey']

    async with sirius_sdk.context(**receiver_params):
        # Get receiver endpoint
        receiver_endpoints = await sirius_sdk.endpoints()
        listener = await sirius_sdk.subscribe()

        for vk_to_send in [sender_verkey, None]:
            for vk_to_receive in [receiver_verkey, None]:
                for endpoint in receiver_endpoints:

                    thread_id = uuid.uuid4().hex
                    msg_under_test = sirius_sdk.messaging.Message({
                        '@id': 'trust-ping-message-' + uuid.uuid4().hex,
                        '@type': 'https://didcomm.org/trust_ping/1.0/ping',
                        "comment": "Hi.",
                        '~thread': {"thid": thread_id}
                    })

                    # Allocate DefaultApi than configured for sender side
                    async with sirius_sdk.context(**sender_params):
                        api = APIDefault(crypto=sirius_sdk.Crypto)
                        await api.send(
                            message=msg_under_test, their_vk=vk_to_receive,
                            endpoint=endpoint.address, my_vk=vk_to_send, routing_keys=endpoint.routing_keys
                        )
                    # Pull events and check them
                    if vk_to_receive is None:
                        vk_to_send = None
                    event = await listener.get_one()
                    assert event.message == msg_under_test, f'Error for sender-vk: {vk_to_send}, routing_keys: {endpoint.routing_keys}'
                    assert event.sender_verkey == vk_to_send
                    assert event.recipient_verkey == vk_to_receive


@pytest.mark.asyncio
async def test_send_batched(test_suite: ServerTestSuite):
    sender_params = test_suite.get_agent_params('agent1')
    receiver1_params = test_suite.get_agent_params('agent2')
    receiver2_params = test_suite.get_agent_params('agent2')
    sender_verkey = list(sender_params['entities'].items())[0][1]['verkey']
    receiver1_verkey = list(receiver1_params['entities'].items())[0][1]['verkey']
    receiver2_verkey = list(receiver2_params['entities'].items())[0][1]['verkey']

    async with sirius_sdk.context(**receiver1_params):
        # Get receiver-1 default endpoint without routing keys
        endpoints = await sirius_sdk.endpoints()
        receiver1_endpoint = [e for e in endpoints if e.routing_keys == []][0]
    async with sirius_sdk.context(**receiver2_params):
        # Get receiver-1 default endpoint without routing keys
        endpoints = await sirius_sdk.endpoints()
        receiver2_endpoint = [e for e in endpoints if e.routing_keys == []][0]

    async def pull_first_event(**sdk):
        async with sirius_sdk.context(**sdk):
            listener = await sirius_sdk.subscribe()
            event = await listener.get_one(timeout=15)
            return event

    listener1_fut = asyncio.ensure_future(pull_first_event(**receiver1_params))
    listener2_fut = asyncio.ensure_future(pull_first_event(**receiver2_params))

    msg_under_test = sirius_sdk.messaging.Message({
        '@id': 'trust-ping-message-' + uuid.uuid4().hex,
        '@type': 'https://didcomm.org/trust_ping/1.0/ping',
        "comment": "Hi.",
        '~thread': {"thid": uuid.uuid4().hex}
    })

    async with sirius_sdk.context(**sender_params):
        api = APIDefault(crypto=sirius_sdk.Crypto)
        batches = [
            sirius_sdk.RoutingBatch(
                their_vk=receiver1_verkey,
                endpoint=receiver1_endpoint.address,
                my_vk=sender_verkey,
                routing_keys=receiver1_endpoint.routing_keys
            ),
            sirius_sdk.RoutingBatch(
                their_vk=receiver2_verkey,
                endpoint=receiver2_endpoint.address,
                my_vk=sender_verkey,
                routing_keys=receiver2_endpoint.routing_keys
            )
        ]
        sender_results = await api.send_batched(message=msg_under_test, batches=batches)
        assert len(sender_results) == 2
        for success, body in sender_results:
            assert success is True
            assert isinstance(body, str)

        listeners_results = await run_coroutines(listener1_fut, listener2_fut)
        assert all([event.message == msg_under_test for event in listeners_results])


import logging
import os
import uuid
import sys
import json
import asyncio

import sirius_sdk

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from consts import AGENT1 as CFG


sirius_sdk.init(**CFG)
BASE = os.path.abspath(os.path.join(os.path.dirname(__file__)))
SESSIONS_FILE = os.path.join(BASE, '.session')


async def run():
    # 1. Try to load old session
    p2p = None
    if os.path.exists(SESSIONS_FILE):
        try:
            with open(SESSIONS_FILE, 'r') as f:
                their_did = f.read()
            p2p = await sirius_sdk.PairwiseList.load_for_did(their_did)
        except:
            logging.exception('Exc')
    if p2p is not None:
        print(f'Do you want to restore old session: "{p2p.their.label}: {p2p.their.did}" ?')
        choice = input('press "yes" if agree: ')
        if choice == 'yes':
            print(f'Session "{p2p.their.did}" was successfully restored')
        else:
            p2p = None

    # 2. Establish connection
    if p2p is None:
        my_did, my_vk = await sirius_sdk.DID.create_and_store_my_did()
        print("=" * 32)
        inv_url = input('Pass invitation URL here: ')
        p2p = await sirius_sdk.recipes.accept_invitation(
            url=inv_url, me=sirius_sdk.Pairwise.Me(my_did, my_vk), my_label='Confidential Storage Provider'
        )
        await sirius_sdk.PairwiseList.ensure_exists(p2p)
        with open(SESSIONS_FILE, 'w+') as f:
            f.truncate(0)
            f.write(p2p.their.did)
        print("!" * 32)
        print(f'Connection with "{p2p.their.label}" was established successfully')
    # Authorize with High level of permission
    auth = sirius_sdk.aries_rfc.ConfidentialStorageAuthProvider()
    await auth.authorize(entity=p2p)
    # For example we will provide 2 types of Vaults: encrypted and non-encrypted
    encrypted_mount_to = os.path.join(os.curdir, '.encrypted')
    non_enc_mount_to = os.path.join(os.curdir, '.non_encrypted')
    if not os.path.exists(encrypted_mount_to):
        os.makedirs(encrypted_mount_to)
    if not os.path.exists(non_enc_mount_to):
        os.makedirs(non_enc_mount_to)
    # Create Vaults
    vault_encrypted = sirius_sdk.recipes.confidential_storage.SimpleDataVault(
        mounted_dir=encrypted_mount_to,
        auth=auth,
        cfg=sirius_sdk.aries_rfc.VaultConfig(
            id=f'did:edvs:{p2p.their.did}#encrypted',
            reference_id='encrypted-vault',
            sequence=1,
            key_agreement=sirius_sdk.aries_rfc.VaultConfig.KeyAgreement(
                id=p2p.me.verkey
            ),
            controller=f'did:peer:{p2p.me.did}',
            delegator=f'did:peer:{p2p.their.did}'
        )
    )
    vault_non_encrypted = sirius_sdk.recipes.confidential_storage.SimpleDataVault(
        mounted_dir=non_enc_mount_to,
        auth=auth,
        cfg=sirius_sdk.aries_rfc.VaultConfig(
            id=f'did:edvs:{p2p.me.did}',
            reference_id='vault',
            sequence=2,
            controller=f'did:peer:{p2p.me.did}',
            delegator=f'did:peer:{p2p.their.did}'
        )
    )
    vault_non_encrypted.cfg.key_agreement = None
    # Run scheduler
    await sirius_sdk.recipes.confidential_storage.schedule_vaults(
        p2p=p2p, vaults=[vault_encrypted, vault_non_encrypted]
    )


if __name__ == '__main__':
    asyncio.get_event_loop().run_until_complete(run())


import json
import datetime
from typing import Optional

import sirius_sdk


PERSIST_STORAGE_TYPE = 'confidential.storage.persist'


def datetime_to_utc_str(tm: datetime.datetime) -> str:
    utc = tm.astimezone(datetime.timezone.utc)
    return utc.isoformat(sep=' ') + 'Z'


async def update_persist_record(key: str, data: dict):
    stored = await ensure_persist_record_exists(key, data)
    if data != stored:
        await sirius_sdk.NonSecrets.update_wallet_record_value(
            type_=PERSIST_STORAGE_TYPE, id_=key, value=json.dumps(data)
        )


async def persist_record_exists(key: str) -> bool:
    opts = sirius_sdk.NonSecretsRetrieveRecordOptions()
    try:
        await sirius_sdk.NonSecrets.get_wallet_record(
            type_=PERSIST_STORAGE_TYPE, id_=key, options=opts
        )
        return True
    except:
        return False


async def store_persist_record_value(key: str, value: str):
    exists = await persist_record_exists(key)
    if exists:
        await sirius_sdk.NonSecrets.update_wallet_record_value(
            type_=PERSIST_STORAGE_TYPE, id_=key, value=value
        )
    else:
        await sirius_sdk.NonSecrets.add_wallet_record(
            type_=PERSIST_STORAGE_TYPE, id_=key, value=value
        )


async def load_persist_record_value(key: str) -> Optional[str]:
    exists = await persist_record_exists(key)
    if exists:
        opts = sirius_sdk.NonSecretsRetrieveRecordOptions()
        opts.retrieve_value = True
        rec = await sirius_sdk.NonSecrets.get_wallet_record(
            type_=PERSIST_STORAGE_TYPE, id_=key, options=opts
        )
        return rec['value']
    else:
        return None


async def ensure_persist_record_exists(key: str, default: dict = None) -> Optional[dict]:
    if default is None:
        default = {}
    opts = sirius_sdk.NonSecretsRetrieveRecordOptions()
    opts.retrieve_value = True
    try:
        rec = await sirius_sdk.NonSecrets.get_wallet_record(
            type_=PERSIST_STORAGE_TYPE, id_=key, options=opts
        )
        states = json.loads(rec['value'])
        return states
    except:
        rec = None
    if rec is None:
        await sirius_sdk.NonSecrets.add_wallet_record(
            type_=PERSIST_STORAGE_TYPE, id_=key, value=json.dumps(default)
        )
        return default


async def ensure_persist_record_missing(key: str):
    try:
        await sirius_sdk.NonSecrets.delete_wallet_record(
            type_=PERSIST_STORAGE_TYPE, id_=key
        )
    except:
        pass



import asyncio
import logging
from enum import IntEnum
from abc import ABC, abstractmethod
from typing import List, Union, Optional, Any, Tuple
from urllib.parse import urlparse

from multipledispatch import dispatch

from sirius_sdk.messaging import Message
from sirius_sdk.errors.exceptions import SiriusConnectionClosed
from sirius_sdk.abstract.bus import AbstractBus
from sirius_sdk.abstract.api import APIContents, APICoProtocols, APINetworks, APIRouter, APIDistributedLocks, \
    APITransport, APICrypto
from sirius_sdk.abstract.batching import RoutingBatch
from sirius_sdk.encryption import P2PConnection
from sirius_sdk.abstract.storage import AbstractImmutableCollection
from sirius_sdk.agent.listener import Listener
from sirius_sdk.agent.wallet.wallets import DynamicWallet
from sirius_sdk.agent.dkms import DKMS
from sirius_sdk.agent.pairwise import AbstractPairwiseList, WalletPairwiseList
from sirius_sdk.agent.storages import InWalletImmutableCollection
from sirius_sdk.agent.microledgers.abstract import AbstractMicroledgerList
from sirius_sdk.agent.microledgers.impl import MicroledgerList
from sirius_sdk.agent.connections import AgentRPC, AgentEvents, BaseAgentConnection
from sirius_sdk.abstract.p2p import Endpoint, TheirEndpoint, Pairwise

from .bus import RpcBus


class SpawnStrategy(IntEnum):
    PARALLEL = 1
    CONCURRENT = 2


class Agent(APIContents, APICoProtocols, APINetworks, APIRouter, APITransport, APIDistributedLocks):
    """
    Agent connection in the self-sovereign identity ecosystem.

    Managing an identity is complex. It is implementation of tools to help you to develop SSI Smart-Contracts logic.
    See details:
      - https://github.com/hyperledger/aries-rfcs/tree/master/concepts/0004-agents
    """

    def __init__(
            self, server_address: str, credentials: bytes,
            p2p: P2PConnection, timeout: int = BaseAgentConnection.IO_TIMEOUT,
            loop: asyncio.AbstractEventLoop = None, storage: AbstractImmutableCollection = None,
            name: str = None, spawn_strategy: SpawnStrategy = SpawnStrategy.PARALLEL,
            external_crypto: APICrypto = None
    ):
        """
        :param server_address: example https://my-cloud-provider.com
        :param credentials: credentials that point websocket connection to your agent and server-side services like
          routing keys maintenance ant etc.
        :param p2p: encrypted connection to establish tunnel to Agent that is running on server-side
        """
        parsed = urlparse(server_address)
        if parsed.scheme not in ['https']:
            logging.warning('Endpoints has non secure scheme, you will have issues for Android/iOS devices')
        self.__server_address = server_address
        self.__credentials = credentials
        self.__p2p = p2p
        self.__rpc = None
        self.__events = None
        self.__wallet = None
        self.__timeout = timeout
        self.__loop = loop
        self.__endpoints = []
        self.__ledgers = {}
        self.__storage = storage
        self.__pairwise_list = None
        self.__microledgers = None
        self.__name = name
        self.__spawn_strategy = spawn_strategy
        self.__external_crypto_service = external_crypto
        self.__bus: Optional[AbstractBus] = None

    @property
    def name(self) -> Optional[str]:
        return self.__name

    @property
    def is_open(self) -> bool:
        return self.__rpc is not None and self.__rpc.is_open

    @property
    def wallet(self) -> DynamicWallet:
        """Indy wallet keys/schemas/CredDefs maintenance"""
        self.__check_is_open()
        return self.__wallet

    def dkms(self, name: str) -> Optional[DKMS]:
        self.__check_is_open()
        return self.__ledgers.get(name, None)

    @property
    def endpoints(self) -> List[Endpoint]:
        self.__check_is_open()
        return self.__endpoints

    @property
    def microledgers(self) -> Optional[AbstractMicroledgerList]:
        self.__check_is_open()
        return self.__microledgers

    @property
    def pairwise_list(self) -> AbstractPairwiseList:
        self.__check_is_open()
        return self.__pairwise_list

    @property
    def bus(self) -> Optional[AbstractBus]:
        return self.__bus

    async def spawn_coprotocol(self) -> AbstractBus:
        if self.__spawn_strategy == SpawnStrategy.PARALLEL:
            rpc = await AgentRPC.create(
                self.__server_address, self.__credentials, self.__p2p,
                self.__timeout, self.__loop, self.__external_crypto_service
            )
        else:
            rpc = self.__rpc
        bus = RpcBus(connector=rpc.connector, p2p=self.__p2p)
        return bus

    async def open(self):
        self.__rpc = await AgentRPC.create(
            self.__server_address, self.__credentials, self.__p2p,
            self.__timeout, self.__loop, self.__external_crypto_service
        )
        self.__bus = RpcBus(connector=self.__rpc.connector, p2p=self.__p2p)
        self.__endpoints = self.__rpc.endpoints
        self.__wallet = DynamicWallet(rpc=self.__rpc)
        if self.__storage is None:
            self.__storage = InWalletImmutableCollection(self.__wallet.non_secrets)
        for network in self.__rpc.networks:
            self.__ledgers[network] = DKMS(
                name=network, api=self.__wallet.ledger,
                issuer=self.__wallet.anoncreds, cache=self.__wallet.cache, storage=self.__storage
            )
        self.__pairwise_list = WalletPairwiseList(api=(self.__wallet.pairwise, self.__wallet.did))
        self.__microledgers = MicroledgerList(api=self.__rpc)

    async def subscribe(self, group_id: str = None) -> Listener:
        self.__check_is_open()
        if group_id:
            extra = {'group_id': group_id}
        else:
            extra = None
        self.__events = await AgentEvents.create(
            self.__server_address, self.__credentials, self.__p2p,
            self.__timeout, self.__loop, self.__external_crypto_service, extra
        )
        return Listener(self.__events, self.pairwise_list)

    async def close(self):
        if self.__rpc:
            await self.__rpc.close()
        if self.__events:
            await self.__events.close()
        self.__wallet = None

    async def ping(self) -> bool:
        success = await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/ping_agent'
        )
        return success

    async def send(
            self, message: Message, their_vk: Union[List[str], str],
            endpoint: str, my_vk: Optional[str] = None, routing_keys: Optional[List[str]] = None
    ) -> (bool, Message):
        """
        Implementation of basicmessage feature
        See details:
          - https://github.com/hyperledger/aries-rfcs/tree/master/features/0095-basic-message

        :param message: Message
          See details:
           - https://github.com/hyperledger/aries-rfcs/tree/master/concepts/0020-message-types
        :param their_vk: Verkey of recipient
        :param endpoint: Endpoint address of recipient
        :param my_vk: VerKey of Sender (AuthCrypt mode)
          See details:
           - https://github.com/hyperledger/aries-rfcs/tree/master/features/0019-encryption-envelope#authcrypt-mode-vs-anoncrypt-mode
        :param routing_keys: Routing key of recipient
        """
        self.__check_is_open()
        await self.__rpc.send_message(
            message=message, their_vk=their_vk, endpoint=endpoint,
            my_vk=my_vk, routing_keys=routing_keys, coprotocol=False
        )

    async def send_to(self, message: Message, to: Pairwise):
        """Implementation of basicmessage feature
        See details:
          - https://github.com/hyperledger/aries-rfcs/tree/master/features/0095-basic-message

        :param message: Message
          See details:
           - https://github.com/hyperledger/aries-rfcs/tree/master/concepts/0020-message-types
        :param to: Pairwise (P2P) connection that have been established outside
        """
        self.__check_is_open()
        await self.send(
            message=message,
            their_vk=to.their.verkey,
            endpoint=to.their.endpoint,
            my_vk=to.me.verkey,
            routing_keys=to.their.routing_keys
        )

    async def send_batched(self, message: Message, batches: List[RoutingBatch]) -> List[Tuple[bool, str]]:
        self.__check_is_open()
        results = await self.__rpc.send_message_batched(message, batches)
        return results

    async def generate_qr_code(self, value: str) -> str:
        """Service for QR codes generation

        You may create PNG image for QR code to share it on Web or others.
        """
        self.__check_is_open()
        resp = await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/admin/1.0/generate_qr',
            params={
                'value': value
            }
        )
        return resp['url']

    async def acquire(self, resources: List[str], lock_timeout: float, enter_timeout: float = 3) -> (bool, List[str]):
        """Acquire N resources given by names

        :param resources: names of resources that you are going to lock
        :param lock_timeout: max timeout resources will be locked. Resources will be automatically unlocked on expire
        :param enter_timeout: timeout to wait resources are released
        """

        self.__check_is_open()
        success, busy = await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/admin/1.0/acquire',
            params={
                'names': resources,
                'enter_timeout': enter_timeout,
                'lock_timeout': lock_timeout
            }
        )
        return success, busy

    async def release(self):
        """Release earlier locked resources"""

        await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/admin/1.0/release'
        )

    async def get_endpoints(self) -> List[Endpoint]:
        return self.endpoints

    async def reopen(self, kill_tasks: bool = False):
        self.__check_is_open()
        await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/reopen',
            params={
                'kill_tasks': kill_tasks
            }
        )

    async def echo(self, message: Any, data: Optional[Any] = None) -> Any:
        self.__check_is_open()
        params = {
            'message': message
        }
        if data:
            params['data'] = data
        ret = await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/admin/1.0/echo',
            params=params
        )
        return ret

    def __check_is_open(self):
        if self.__rpc and self.__rpc.is_open:
            return self.__endpoints
        else:
            raise SiriusConnectionClosed('Open Agent at first!')



import pytest

import pytest

import sirius_sdk
from sirius_sdk.errors.exceptions import SiriusCryptoError
from sirius_sdk.hub.defaults.default_crypto import DefaultCryptoService

from ..helpers import ServerTestSuite


@pytest.mark.asyncio
async def test_sane():
    crypto_under_test = DefaultCryptoService()
    # check key creation
    vk1 = await crypto_under_test.create_key()
    vk2 = await crypto_under_test.create_key()
    assert vk1 != vk2
    seed = '0'*32
    vk3 = await crypto_under_test.create_key(seed)
    vk4 = await crypto_under_test.create_key(seed)
    assert vk3 == vk4
    # Check metadata
    metadata = {'key': 'value'}
    await crypto_under_test.set_key_metadata(vk1, metadata)
    m1 = await crypto_under_test.get_key_metadata(vk1)
    assert m1 == metadata
    m2 = await crypto_under_test.get_key_metadata(vk2)
    assert m2 is None
    # check errors
    unknown_vk = 'A' * 44
    with pytest.raises(SiriusCryptoError):
        await crypto_under_test.get_key_metadata(unknown_vk)


@pytest.mark.asyncio
async def test_pack_messages():
    sender = DefaultCryptoService()
    receiver = DefaultCryptoService()
    vk_sender = await sender.create_key()
    vk_receiver = await receiver.create_key()

    message = 'Test-Message'

    packed = await sender.pack_message(
        message=message, recipient_verkeys=[vk_receiver], sender_verkey=vk_sender
    )
    decrypted = await receiver.unpack_message(packed)
    assert decrypted['message'] == message
    assert decrypted['recipient_verkey'] == vk_receiver
    assert decrypted['sender_verkey'] == vk_sender

    alien_key = 'BntaRZtrnqHE6WYNgJCmcqC3kwwfFrvgXwdTJumywcGc'
    packed = await sender.pack_message(
        message=message, recipient_verkeys=[alien_key], sender_verkey=vk_sender
    )
    with pytest.raises(SiriusCryptoError):
        await receiver.unpack_message(packed)


@pytest.mark.asyncio
async def test_anon_crypt(test_suite: ServerTestSuite):
    cloud = test_suite.get_agent_params('agent1')
    crypto_under_test = DefaultCryptoService()
    recipient_vk = await crypto_under_test.create_key()
    message = b'Test-Message'

    async with sirius_sdk.context(**cloud):
        encrypted = await sirius_sdk.Crypto.anon_crypt(recipient_vk, message)

    decrypted = await crypto_under_test.anon_decrypt(recipient_vk, encrypted)
    assert decrypted == message


@pytest.mark.asyncio
async def test_anon_decrypt(test_suite: ServerTestSuite):
    cloud = test_suite.get_agent_params('agent1')
    crypto_under_test = DefaultCryptoService()
    message = b'Test-Message'

    async with sirius_sdk.context(**cloud):
        recip_vk = await sirius_sdk.Crypto.create_key()

    encrypted = await crypto_under_test.anon_crypt(recip_vk, message)

    async with sirius_sdk.context(**cloud):
        decrypted = await sirius_sdk.Crypto.anon_decrypt(recip_vk, encrypted)

    assert decrypted == message



import base64
import dataclasses
import datetime
import logging
from enum import Enum
from dataclasses import dataclass
from typing import Optional, List, Any, Union

from pytime import pytime
from sirius_sdk.base import Message
from .decorators import *


def parse_datetime(val: str) -> Optional[datetime.datetime]:
    orig_val = val
    if val:
        if val[-1] == 'Z' and len(val.split(' ')) == 2:
            val = val[:-1].replace(' ', 'T')
        for func in [pytime.parse, datetime.datetime.fromisoformat]:
            try:
                return func(val)
            except Exception as e:
                pass
        logging.warning(f'Error while parsing string formatted datetime "{orig_val}"')
        return None
    else:
        return None


class PleaseAckMixin:
    """Explains how one party can request an acknowledgment to and clarify the status of processes.
       - RFC Aries: https://github.com/hyperledger/aries-rfcs/tree/main/features/0317-please-ack
    """

    @property
    def ack_message_id(self) -> Optional[str]:
        return self.get_ack_message_id(self)

    @property
    def please_ack(self) -> bool:
        return self.get(PLEASE_ACK_DECORATOR, None) is not None

    @please_ack.setter
    def please_ack(self, flag: bool):
        if flag:
            self[PLEASE_ACK_DECORATOR] = {'message_id': self.id}
        elif PLEASE_ACK_DECORATOR in self:
            del self[PLEASE_ACK_DECORATOR]

    @staticmethod
    def get_ack_message_id(message: Message) -> Optional[str]:
        return message.get(PLEASE_ACK_DECORATOR, {}).get('message_id', None) or message.id


class ThreadMixin:
    """Definition of the message @id field and the ~thread decorator

      - RFC Aries: https://github.com/hyperledger/aries-rfcs/tree/main/concepts/0008-message-id-and-threading
    """

    @dataclass
    class Thread:
        thid: str = None
        pthid: str = None
        sender_order: int = None
        received_orders: dict = None

        @property
        def is_filled(self) -> bool:
            return self.thid is not None or \
                   self.pthid is not None or \
                   self.sender_order is not None or \
                   self.received_orders is not None

    @property
    def thread_id(self) -> Optional[str]:
        return self.get(THREAD_DECORATOR, {}).get('thid', None)

    @thread_id.setter
    def thread_id(self, thid: str):
        thread = self.get(THREAD_DECORATOR, {})
        thread['thid'] = thid
        self[THREAD_DECORATOR] = thread

    @property
    def thread(self) -> Optional[Thread]:
        return self.get_thread(self)

    @thread.setter
    def thread(self, value: Thread = None):
        self.set_thread(self, value)

    @staticmethod
    def get_thread(message: Message) -> Optional[Thread]:
        d = message.get(THREAD_DECORATOR, {})
        thid = d.get('thid', None)
        pthid = d.get('pthid', None)
        if 'sender_order' in d:
            try:
                sender_order = int(d['sender_order'])
            except ValueError:
                sender_order = None
        else:
            sender_order = None
        if 'received_orders' in d:
            received_orders = d['received_orders']
            if not isinstance(received_orders, dict):
                received_orders = None
        else:
            received_orders = None

        thread = ThreadMixin.Thread(
            thid=thid, pthid=pthid, sender_order=sender_order, received_orders=received_orders
        )
        if thread.is_filled:
            return thread
        else:
            return None

    @staticmethod
    def set_thread(message: Message, value: Thread = None):
        if value is not None and value.is_filled:
            d = {}
            if value.thid:
                d['thid'] = value.thid
            if value.pthid:
                d['pthid'] = value.pthid
            if value.sender_order is not None:
                d['sender_order'] = value.sender_order
            if value.received_orders is not None:
                d['received_orders'] = value.received_orders
            message[THREAD_DECORATOR] = d
        else:
            if THREAD_DECORATOR in self:
                del message[THREAD_DECORATOR]


class Attach(dict):

    def __init__(self, id: str = None, mime_type: str = None, filename: str = None, lastmod_time: str = None,
                 description: str = None, data: bytes = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if id is not None:
            self["@id"] = id
        if mime_type is not None:
            self["mime-type"] = mime_type
        if filename is not None:
            self["filename"] = filename
        if lastmod_time is not None:
            self["lastmod_time"] = lastmod_time
        if description is not None:
            self["description"] = description
        if data is not None:
            self["data"] = {
                "base64": base64.b64encode(data).decode()
            }

    @property
    def id(self) -> Optional[str]:
        return self['@id']

    @property
    def mime_type(self) -> Optional[str]:
        return self['mime-type']

    @property
    def filename(self) -> Optional[str]:
        return self['filename']

    @property
    def lastmod_time(self) -> Optional[str]:
        return self['lastmod_time']

    @property
    def description(self) -> Optional[str]:
        return self['description']

    @property
    def data(self) -> Optional[bytes]:
        return base64.b64decode(self['data']["base64"])


class AttachesMixin:
    """Explains the three canonical ways to attach data to an agent message.
      - RFC Aries: https://github.com/hyperledger/aries-rfcs/tree/master/concepts/0017-attachments
    """

    @property
    def attaches(self) -> List[Attach]:
        if "~attach" in self:
            return self["~attach"]
        else:
            return []

    def add_attach(self, att: Attach):
        if "~attach" not in self:
            self["~attach"] = []
        self["~attach"] += [att]


class ReturnRouteMixin:
    """Agents can indicate that an inbound message transmission may also be used as a return route for messages.
       This allows for transports of increased efficiency as well as agents without an inbound route.

      - RFC Aries(v1): https://github.com/hyperledger/aries-rfcs/tree/main/features/0092-transport-return-route
      - RFC DIDComm(v2): https://github.com/decentralized-identity/didcomm-messaging/blob/main/extensions/return_route/main.md
    """

    class RouteType(Enum):
        NONE = 'none'
        ALL = 'all'
        THREAD = 'thread'

    @property
    def return_route(self) -> RouteType:
        return self.get_return_route(self)

    @return_route.setter
    def return_route(self, value: RouteType):
        self.set_return_route(self, value)

    @classmethod
    def get_return_route(cls, message: Message) -> RouteType:
        value = message.get(TRANSPORT_DECORATOR, {}).get('return_route', None)
        if value is None:
            return cls.RouteType.NONE
        else:
            if value == cls.RouteType.ALL.value:
                return cls.RouteType.ALL
            elif value == cls.RouteType.THREAD.value:
                return cls.RouteType.THREAD
            else:
                return cls.RouteType.NONE

    @classmethod
    def set_return_route(cls, message: Message, transport: RouteType):
        message[TRANSPORT_DECORATOR] = {
            'return_route': transport.value
        }


class TimingMixin:
    """Explain how timing of agent messages can be communicated and constrained.

      - RFC Aries: https://github.com/hyperledger/aries-rfcs/tree/main/features/0032-message-timing
    """

    @dataclass
    class Timing:
        # The timestamp when the preceding message in this thread (the one that elicited this message as a response)
        # was received. Or, on a dynamically composed forward message, the timestamp when an upstream relay first
        # received the message it's now asking to be forwarded.
        in_time: datetime.datetime = None

        # The timestamp when the message was emitted
        out_time: datetime.datetime = None

        #  Ideally, the decorated message should be processed by the the specified timestamp
        stale_time: datetime.datetime = None

        # The decorated message should be considered invalid or expired if encountered after the specified timestamp
        expires_time: datetime.datetime = None

        # Wait at least this many milliseconds before processing the message
        delay_milli: int = None

        # Wait until this time before processing the message
        wait_until_time: datetime.datetime = None

        def __init__(
                self, in_time: Union[str, datetime.datetime] = None, out_time: Union[str, datetime.datetime] = None,
                stale_time: Union[str, datetime.datetime] = None, expires_time: Union[str, datetime.datetime] = None,
                delay_milli: int = None, wait_until_time: Union[str, datetime.datetime] = None, **kwargs
        ):
            self.in_time = self.__read_value_safe('in_time', in_time)
            self.out_time = self.__read_value_safe('out_time', out_time)
            self.stale_time = self.__read_value_safe('stale_time', stale_time)
            self.expires_time = self.__read_value_safe('expires_time', expires_time)
            self.delay_milli = self.__read_value_safe('delay_milli', delay_milli)
            self.wait_until_time = self.__read_value_safe('wait_until_time', wait_until_time)

        @property
        def is_filled(self) -> bool:
            return any(
                [t is not None for t in (self.in_time, self.out_time, self.stale_time, self.expires_time, self.delay_milli, self.wait_until_time)]
            )

        def to_json(self) -> dict:
            obj = dataclasses.asdict(self)
            return {key: obj[key] for key, value in obj.items() if value is not None}

        def create_from_json(self, js: dict) -> 'TimingMixin.Timing':
            upd_kwargs = {}
            for key, value in js.items():
                safe_val = self.__read_value_safe(key, value)
                upd_kwargs[key] = safe_val
            return dataclasses.replace(self, **upd_kwargs)

        def __read_value_safe(self, name: str, value: Any) -> Any:
            fields = dataclasses.fields(self)
            actual_fields = list(filter(lambda f: f.name == name, fields))
            if actual_fields:
                field = actual_fields[0]
                if field.type == type(value):
                    return value
                else:
                    if field.type == int:
                        if isinstance(value, str) and value.isdigit():
                            return int(value)
                        else:
                            return None
                    elif field.type == datetime.datetime:
                        if isinstance(value, str):
                            return parse_datetime(value)
                        else:
                            return None
                    else:
                        return None
            else:
                return None

    @property
    def timing(self) -> Timing:
        return self.get_timing(self)

    @timing.setter
    def timing(self, value: Timing):
        self.set_timing(self, value)

    @classmethod
    def get_timing(cls, message: Message) -> Optional[Timing]:
        js = message.get(TIMING_DECORATOR, {})
        if js:
            value = TimingMixin.Timing(**js)
            return value if value.is_filled else None
        else:
            return None

    @classmethod
    def set_timing(cls, message: Message, value: Timing):
        js = value.to_json() if value else {}
        if js:
            message[TIMING_DECORATOR] = js
        elif TIMING_DECORATOR in message.keys():
            del message[TIMING_DECORATOR]



import uuid

import pytest

import sirius_sdk
from sirius_sdk import Agent
from sirius_sdk.agent.aries_rfc.feature_0482_coprotocol import Caller, Called

from tests.conftest import get_pairwise
from tests.helpers import ServerTestSuite, run_coroutines
from sirius_sdk.agent.aries_rfc.feature_0482_coprotocol.messages import *
from sirius_sdk.recipes import as_caller, as_called


@pytest.mark.asyncio
async def test_invoke(A: Agent, B: Agent, test_suite: ServerTestSuite):
    caller = A
    called = B
    caller_params = test_suite.get_agent_params('agent1')
    called_params = test_suite.get_agent_params('agent2')
    await caller.open()
    await called.open()
    try:
        caller_2_called = await get_pairwise(caller, called)
        test_log = {}

        async def run_caller():
            uri, cred, p2p = caller_params['server_address'], caller_params['credentials'], caller_params['p2p']
            cfg = sirius_sdk.Config().setup_cloud(uri, cred, p2p)
            async with sirius_sdk.context(cfg):
                machine = Caller(called=caller_2_called, thid=uuid.uuid4().hex)
                assert machine.state == Caller.State.NULL
                success, ctx = await machine.bind(
                    cast=[
                        {'role': 'payer', 'id': None},
                        {'role': 'provider', 'id': caller_2_called.their.did}
                    ]
                )
                assert machine.state == Caller.State.ATTACHED
                test_log['caller_attached'] = (success, dict(**ctx))
                await machine.input(data={'key': 'value'}, extra_key='extra-value')
                success, data, extra = await machine.wait_output()
                test_log['caller_output_1'] = (success, dict(**data), dict(**extra))
                await machine.detach()
                assert machine.state == Caller.State.DETACHED
                test_log['caller_done'] = True

        async def run_called():
            uri, cred, p2p = called_params['server_address'], called_params['credentials'], called_params['p2p']
            cfg = sirius_sdk.Config().setup_cloud(uri, cred, p2p)
            async with sirius_sdk.context(cfg):
                listener = await sirius_sdk.subscribe()
                async for event in listener:
                    request = event.message
                    if isinstance(request, CoProtocolBind):
                        machine = await Called.open(caller=event.pairwise, request=request)
                        assert machine.state == Called.State.NULL
                        test_log['called_bind_context'] = dict(**machine.context)
                        assert machine.state == Called.State.NULL
                        await machine.attach(field1='value-1', field2='value-2')
                        assert machine.state == Called.State.ATTACHED
                        success, data, extra = await machine.wait_input()
                        test_log['called_input_1'] = (success, dict(**data), dict(**extra))
                        await machine.output(data={'key': 'value-from-called'}, extra_key='extra-value-from-called')
                        try:
                            await machine.wait_input()
                        except Called.CoProtocolDetachedByCaller as e:
                            test_log['called_input_exception'] = str(e)
                        assert machine.state == Called.State.DONE
                        test_log['called_done'] = True
                        return
        print('> begin')
        await run_coroutines(run_caller(), run_called(), timeout=5)
        print('> end')
        assert test_log.get('called_bind_context') == {'co_binding_id': None, 'cast': [{'role': 'payer', 'id': None}, {'role': 'provider', 'id': 'T8MtAB98aCkgNLtNfQx6WG'}]}
        assert test_log.get('caller_attached') == (True, {'field1': 'value-1', 'field2': 'value-2'})
        assert test_log.get('called_input_1') == (True, {'key': 'value'}, {'extra_key': 'extra-value'})
        assert test_log.get('caller_output_1') == (True, {'key': 'value-from-called'}, {'extra_key': 'extra-value-from-called'})
        assert test_log.get('called_input_exception') is not None
        assert test_log.get('called_done') is True
        assert test_log.get('caller_done') is True
    finally:
        await caller.close()
        await called.close()


@pytest.mark.asyncio
async def test_caller_problem_report(A: Agent, B: Agent, test_suite: ServerTestSuite):
    caller = A
    called = B
    caller_params = test_suite.get_agent_params('agent1')
    called_params = test_suite.get_agent_params('agent2')
    await caller.open()
    await called.open()
    try:
        caller_2_called = await get_pairwise(caller, called)
        test_log = {}

        async def run_caller():
            uri, cred, p2p = caller_params['server_address'], caller_params['credentials'], caller_params['p2p']
            cfg = sirius_sdk.Config().setup_cloud(uri, cred, p2p)
            async with sirius_sdk.context(cfg):
                machine = Caller(called=caller_2_called, thid=uuid.uuid4().hex)
                success, ctx = await machine.bind(
                    cast=[
                        {'role': 'payer', 'id': None},
                        {'role': 'provider', 'id': caller_2_called.their.did}
                    ]
                )
                assert success is True
                await machine.raise_problem(problem_code='123', explain='TEST')
                assert machine.state == Caller.State.ATTACHED
                test_log['caller_done'] = True

        async def run_called():
            uri, cred, p2p = called_params['server_address'], called_params['credentials'], called_params['p2p']
            cfg = sirius_sdk.Config().setup_cloud(uri, cred, p2p)
            async with sirius_sdk.context(cfg):
                listener = await sirius_sdk.subscribe()
                async for event in listener:
                    request = event.message
                    if isinstance(request, CoProtocolBind):
                        machine = await Called.open(caller=event.pairwise, request=request)
                        await machine.attach()
                        success, data, extra = await machine.wait_input()
                        test_log['called_input_1'] = success, data, extra
                        if success is False:
                            test_log['called_problem_report'] = (machine.problem_report.problem_code, machine.problem_report.explain)
                        assert machine.state == Called.State.ATTACHED
                        test_log['called_done'] = True
                        return
        print('> begin')
        await run_coroutines(run_caller(), run_called(), timeout=5)
        print('> end')
        assert test_log.get('called_input_1') == (False, None, None)
        assert test_log.get('called_problem_report') == ('123', 'TEST')
        assert test_log.get('called_done') is True
        assert test_log.get('caller_done') is True
    finally:
        await caller.close()
        await called.close()


@pytest.mark.asyncio
async def test_recipe_1(A: Agent, B: Agent, test_suite: ServerTestSuite):
    caller = A
    called = B
    caller_params = test_suite.get_agent_params('agent1')
    called_params = test_suite.get_agent_params('agent2')
    await caller.open()
    await called.open()
    try:
        caller_2_called = await get_pairwise(caller, called)
        test_log = {}

        async def run_caller():
            uri, cred, p2p = caller_params['server_address'], caller_params['credentials'], caller_params['p2p']
            cfg = sirius_sdk.Config().setup_cloud(uri, cred, p2p)
            async with sirius_sdk.context(cfg):
                async with as_caller(called=caller_2_called, thid=uuid.uuid4().hex, cast={'cast-key': 'cast-val'}) as proto:
                    assert proto.state == Caller.State.ATTACHED
                    test_log['caller_attached'] = True
                    await proto.input(data={'key': 'value'}, extra_key='extra-value')
                    success, data, extra = await proto.wait_output()
                    test_log['caller_output_1'] = (success, dict(**data), dict(**extra))
                    await proto.detach()
                    assert proto.state == Caller.State.DETACHED
                    test_log['caller_done'] = True

        async def run_called():
            uri, cred, p2p = called_params['server_address'], called_params['credentials'], called_params['p2p']
            cfg = sirius_sdk.Config().setup_cloud(uri, cred, p2p)
            async with sirius_sdk.context(cfg):
                listener = await sirius_sdk.subscribe()
                async for event in listener:
                    request = event.message
                    if isinstance(request, CoProtocolBind):
                        async with as_called(caller=event.pairwise, request=request) as proto:
                            await proto.attach()
                            success, data, extra = await proto.wait_input()
                            test_log['called_input_1'] = success, data, extra
                            await proto.output(data={'key': 'value-from-called'}, extra_key='extra-value-from-called')
                            try:
                                await proto.wait_input()
                            except Called.CoProtocolDetachedByCaller as e:
                                test_log['called_input_exception'] = str(e)
                            assert proto.state == Called.State.DONE
                            test_log['called_done'] = True
                            return
        print('> begin')
        await run_coroutines(run_caller(), run_called(), timeout=5)
        print('> end')
        for event in ['caller_attached', 'called_input_1', 'caller_output_1', 'called_input_exception', 'called_done', 'caller_done']:
            assert event in test_log.keys()
    finally:
        await caller.close()
        await called.close()




import datetime
import uuid

from sirius_sdk.agent.aries_rfc.mixins import *


def test_feature_0032_timing_decorator():
    # Check creation
    now = datetime.datetime.now()
    timing1 = TimingMixin.Timing(
        in_time='2019-01-23 18:03:27.123Z', delay_milli=1000, expires_time=now, any_field='ANY'
    )
    assert timing1.in_time == parse_datetime('2019-01-23 18:03:27.123Z')
    assert timing1.delay_milli == 1000
    assert timing1.expires_time == now
    assert timing1.wait_until_time is None

    # Check to_json
    js = timing1.to_json()
    timing2 = TimingMixin.Timing(**js)
    assert timing1 == timing2

    # Check update
    timing3 = timing2.create_from_json(
        dict(in_time='2019-01-24 00:00Z', delay_milli=2000, any_field='ANY', stale_time='2019-01-24 18:25Z')
    )
    assert timing3.in_time == parse_datetime('2019-01-24 00:00Z')
    assert timing3.delay_milli == 2000
    assert timing3.expires_time == now
    assert timing3.stale_time == parse_datetime('2019-01-24 18:25Z')

    # Check update clear field
    timing4 = timing2.create_from_json(dict(in_time=None))
    assert timing4.in_time is None


def test_feature_0032_timing_message():
    msg = Message({
        '@id': uuid.uuid4().hex,
        '@type': 'https://didcomm.org/protocol/1.0/message-x',
        '~timing': {
            "in_time": "2019-01-23 18:03:27.123Z",
            "out_time": "2019-01-23 18:03:27.123Z",
            "stale_time": "2019-01-24 18:25Z",
            "expires_time": "2019-01-25 18:25Z",
            "delay_milli": 12345,
            "wait_until_time": "2019-01-24 00:00Z"
        }
    })
    # Get
    timing = TimingMixin.get_timing(msg)
    assert timing.in_time == parse_datetime("2019-01-23 18:03:27.123Z")
    assert timing.delay_milli == 12345
    assert timing.wait_until_time == parse_datetime("2019-01-24 00:00Z")

    # Set
    now = datetime.datetime.now()
    TimingMixin.set_timing(msg, TimingMixin.Timing(in_time=now))
    timing = TimingMixin.get_timing(msg)
    assert timing.in_time == now
    assert timing.delay_milli is None
    assert timing.wait_until_time is None

    # Clear
    TimingMixin.set_timing(msg, None)
    assert TIMING_DECORATOR not in msg
    timing = TimingMixin.get_timing(msg)
    assert timing is None



import sirius_sdk


AGENT1 = {
    'server_uri': 'https://demo.socialsirius.com',
    'credentials': 'ez8ucxfrTiV1hPX99MHt/C/MUJCo8OmN4AMVmddE/sew8gBzsOg040FWBSXzHd9hDoj5B5KN4aaLiyzTqkrbD3uaeSwmvxVsqkC0xl5dtIc='.encode(),
    'p2p': sirius_sdk.P2PConnection(
            my_keys=('6QvQ3Y5pPMGNgzvs86N3AQo98pF5WrzM1h6WkKH3dL7f', '28Au6YoU7oPt6YLpbWkzFryhaQbfAcca9KxZEmz22jJaZoKqABc4UJ9vDjNTtmKSn2Axfu8sT52f5Stmt7JD4zzh'),
            their_verkey='6oczQNLU7bSBzVojkGsfAv3CbXagx7QLUL7Yj1Nba9iw'
        )
}


AGENT2 = {
    'server_uri': 'https://demo.socialsirius.com',
    'credentials': 'ez8ucxfrTiV1hPX99MHt/NRtCY78r2bCZO8nJ7ooWxDa6TQbCWUvnpylTJSRnMq3Doj5B5KN4aaLiyzTqkrbDwMKo4RJ3alpnUUd4iyxgqE='.encode(),
    'p2p': sirius_sdk.P2PConnection(
        my_keys=('5o6wXAYT3A8svdog2t4M3gk15iXNW8yvxVu3utJHAD7g', '2xsAzx4URZGY8imWRL5jFAbQqvdFHw4ZbuxxoAADSqVCFTbiwZYhw4gPVA5dsqbJSsLxbac7ath4sFiHYzyVsEDY'),
        their_verkey='GoPE3ZkJXKjrNnP8LdJuzhbxKjAwND8Q9vSosqzTeAJd'
    )
}


AGENT3 = {
    'server_uri': 'https://demo.socialsirius.com',
    'credentials': 'ez8ucxfrTiV1hPX99MHt/P+YgoaBDJV7S03Nxc26pIVlgwkbSZ0XjQ9fEVd4Xrq+Doj5B5KN4aaLiyzTqkrbD8j/KbG7UG4Jfx2kkFcXAvc='.encode(),
    'p2p': sirius_sdk.P2PConnection(
        my_keys=('6RZN88AEYsYQH6WXyunMt8JXLFjenDqrRGeQayG5zY15', '265jVEbBup5EJ9pHJrrRWDnm5rxcZhHkn6FPbcH1su9HMT28yv8BHwithHT8PnFxx91zPVeBiXBvTywqLk3P3vfh'),
        their_verkey='FCzrVZqZbn1PAJQNSqgaLP6DXZhKKUtK5wU3TNnG7d5P'
    )
}


AGENT4 = {
    'server_uri': 'https://demo.socialsirius.com',
    'credentials': 'ez8ucxfrTiV1hPX99MHt/JZL1h63sUO9saQCgn2BsaC2EndwDSYpOo6eFpn8xP8ZDoj5B5KN4aaLiyzTqkrbDxrbAe/+2uObPTl6xZdXMBs='.encode(),
    'p2p': sirius_sdk.P2PConnection(
        my_keys=('B1n1Hwj1USs7z6FAttHCJcqhg7ARe7xtcyfHJCdXoMnC', 'y7fwmKxfatm6SLN6sqy6LFFjKufgzSsmqA2D4WZz55Y8W7JFeA3LvmicC36E8rdHoAiFhZgSf4fuKmimk9QyBec'),
        their_verkey='5NUzoX1YNm5VXsgzudvVikN7VQpRf5rhaTnPxyu12eZC'
    )
}


PROVER_SECRET_ID = 'prover-secret-id'



import json
import aiohttp
import asyncio
import datetime
from abc import ABC, abstractmethod
from typing import List, Union, Optional

from sirius_sdk.abstract.batching import RoutingBatch
from sirius_sdk.abstract.p2p import Endpoint
from sirius_sdk.abstract.api import APICrypto
from sirius_sdk.base import WebSocketConnector
from sirius_sdk.encryption import P2PConnection
from sirius_sdk.rpc import AddressedTunnel, build_request, Future
from sirius_sdk.messaging import Message, Type as MessageType
from sirius_sdk.errors.exceptions import *
from sirius_sdk.messaging.transport import http_send


class BaseAgentConnection(ABC):

    IO_TIMEOUT = 60
    MSG_TYPE_CONTEXT = 'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/context'

    def __init__(
            self, server_address: str, credentials: bytes,
            p2p: P2PConnection, timeout: int = IO_TIMEOUT,
            loop: asyncio.AbstractEventLoop = None, external_crypto: APICrypto = None,
            extra: dict = None
    ):
        self._connector = WebSocketConnector(
            server_address=server_address,
            path=self._path(),
            credentials=credentials,
            timeout=timeout,
            loop=loop,
            extra=extra
        )
        self.__loop = loop or asyncio.get_event_loop()
        self._p2p = p2p
        self._timeout = timeout
        self._external_crypto_service = external_crypto

    def __del__(self):
        if self.__loop and self.__loop.is_running():
            asyncio.ensure_future(self.close())

    @property
    def timeout(self):
        return self._timeout

    @timeout.setter
    def timeout(self, value: int):
        if value is None:
            self._timeout = None
        elif value > 0:
            self._timeout = value
        else:
            raise RuntimeError('Timeout must be > 0')

    @property
    def is_open(self):
        return self._connector.is_open

    async def close(self):
        await self._connector.close()

    @classmethod
    async def create(
            cls, server_address: str, credentials: bytes,
            p2p: P2PConnection, timeout: int = IO_TIMEOUT,
            loop: asyncio.AbstractEventLoop = None, external_crypto: APICrypto = None,
            extra: dict = None
    ):
        instance = cls(server_address, credentials, p2p, timeout, loop, external_crypto, extra)
        await instance._connector.open()
        payload = await instance._connector.read(timeout=timeout)
        context = Message.deserialize(payload.decode())
        msg_type = context.get('@type', None)
        if msg_type is None:
            raise RuntimeError('message @type is empty')
        elif msg_type != cls.MSG_TYPE_CONTEXT:
            raise RuntimeError('message @type is empty')
        else:
            await instance._setup(context)
        return instance

    @classmethod
    @abstractmethod
    def _path(cls):
        raise NotImplemented()

    async def _setup(self, context: Message):
        pass


class AgentRPC(BaseAgentConnection):
    """RPC service.

    Proactive form of Smart-Contract design
    """

    EXPIRATION_OFF = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__tunnel_rpc = None
        self.__tunnel_coprotocols = None
        self.__endpoints = []
        self.__networks = []
        self.__websockets = {}
        self.__prefer_agent_side = True
        self.__tcp_connector = aiohttp.TCPConnector(ssl=False, keepalive_timeout=60)

    @property
    def endpoints(self) -> List[Endpoint]:
        return self.__endpoints

    @property
    def connector(self) -> WebSocketConnector:
        return self._connector

    @property
    def networks(self) -> List[str]:
        return self.__networks

    async def remote_call(
            self, msg_type: str, params: dict = None, wait_response: bool = True, reconnect_on_error: bool = True
    ) -> Any:
        """Call Agent services

        :param msg_type:
        :param params:
        :param wait_response: wait for response
        :param reconnect_on_error: try reconnect if server was closed recources
        :return:
        """
        try:
            if not self._connector.is_open:
                raise SiriusConnectionClosed('Open agent connection at first')
            if self._timeout and not self.EXPIRATION_OFF:
                expiration_time = datetime.datetime.now() + datetime.timedelta(seconds=self._timeout)
            else:
                expiration_time = None
            future = Future(
                tunnel=self.__tunnel_rpc,
                expiration_time=expiration_time
            )
            request = build_request(
                msg_type=msg_type,
                future=future,
                params=params or {}
            )
            msg_typ = MessageType.from_str(msg_type)
            encrypt = msg_typ.protocol not in ['admin', 'microledgers', 'microledgers-batched', 'bus']
            if not await self.__tunnel_rpc.post(message=request, encrypt=encrypt):
                raise SiriusRPCError()
            if wait_response:
                success = await future.wait(timeout=self._timeout)
                if success:
                    if future.has_exception():
                        future.raise_exception()
                    else:
                        return future.get_value()
                else:
                    raise SiriusTimeoutRPC()
        except SiriusConnectionClosed:
            if reconnect_on_error:
                await self._reopen()
                return await self.remote_call(msg_type, params, wait_response, reconnect_on_error=False)
            else:
                raise
        
    async def send_message(
            self, message: Message,
            their_vk: Union[List[str], str], endpoint: str,
            my_vk: Optional[str], routing_keys: Optional[List[str]],
            coprotocol: bool = False, ignore_errors: bool = False
    ) -> Optional[Message]:
        """Send Message to other Indy compatible agent
        
        :param message: message
        :param their_vk: Verkey of recipients
        :param endpoint: Endpoint Address of recipient
        :param my_vk: Verkey of sender (None for anocrypt mode)
        :param routing_keys: Routing keys if it is exists
        :param coprotocol: True if message is part of co-protocol stream
            See:
             - https://github.com/hyperledger/aries-rfcs/tree/master/concepts/0003-protocols
             - https://github.com/hyperledger/aries-rfcs/tree/master/concepts/0008-message-id-and-threading
        :param ignore_errors: bool hide any raised exception
        :return: Response message if coprotocol is True
        """
        if not self._connector.is_open:
            raise SiriusConnectionClosed('Open agent connection at first')
        if isinstance(their_vk, str):
            recipient_verkeys = [their_vk]
        else:
            recipient_verkeys = their_vk
        if self._external_crypto_service:
            wired = await self._external_crypto_service.pack_message(
                message, recipient_verkeys, my_vk
            )
            if self.__prefer_agent_side:
                params = {
                    'message': wired,
                    'timeout': self.timeout,
                    'endpoint_address': endpoint
                }
                ok, body = await self.remote_call(
                    msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/admin/1.0/send_enc_message',
                    params=params
                )
            else:
                if endpoint.startswith('ws://') or endpoint.startswith('wss://'):
                    ws = await self.__get_websocket(endpoint)
                    await ws.send_bytes(wired)
                    ok, body = True, b''
                else:
                    ok, body = await http_send(wired, endpoint, timeout=self.timeout, connector=self.__tcp_connector)
                body = body.decode()
        else:
            params = {
                'message': message,
                'routing_keys': routing_keys or [],
                'recipient_verkeys': recipient_verkeys,
                'sender_verkey': my_vk
            }
            if self.__prefer_agent_side:
                params['timeout'] = self.timeout
                params['endpoint_address'] = endpoint
                ok, body = await self.remote_call(
                    msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/send_message',
                    params=params
                )
            else:
                wired = await self.remote_call(
                    msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/prepare_message_for_send',
                    params=params
                )
                if endpoint.startswith('ws://') or endpoint.startswith('wss://'):
                    ws = await self.__get_websocket(endpoint)
                    await ws.send_bytes(wired)
                    ok, body = True, b''
                else:
                    ok, body = await http_send(wired, endpoint, timeout=self.timeout, connector=self.__tcp_connector)
                body = body.decode()
        if not ok:
            if not ignore_errors:
                raise SiriusRPCError(body)
        else:
            if coprotocol:
                response = await self.read_protocol_message()
                return response
            else:
                return None

    async def send_message_batched(self, message: Message, batches: List[RoutingBatch]) -> List[Any]:
        if not self._connector.is_open:
            raise SiriusConnectionClosed('Open agent connection at first')
        params = {
            'message': message,
            'timeout': self.timeout,
            'batches': batches,
        }
        results = await self.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/send_message_batched',
            params=params
        )
        return results

    async def read_protocol_message(self) -> Message:
        response = await self.__tunnel_coprotocols.receive(timeout=self._timeout)
        return response

    async def start_protocol_with_threading(self, thid: str, ttl: int=None):
        await self.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/start_protocol',
            params={
                'thid': thid,
                'channel_address': self.__tunnel_coprotocols.address,
                'ttl': ttl
            }
        )

    async def start_protocol_with_threads(self, threads: List[str], ttl: int=None):
        await self.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/start_protocol',
            params={
                'threads': threads,
                'channel_address': self.__tunnel_coprotocols.address,
                'ttl': ttl
            }
        )

    async def start_protocol_for_p2p(self, sender_verkey: str, recipient_verkey: str, protocols: List[str], ttl: int=None):
        await self.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/start_protocol',
            params={
                'sender_verkey': sender_verkey,
                'recipient_verkey': recipient_verkey,
                'protocols': protocols,
                'channel_address': self.__tunnel_coprotocols.address,
                'ttl': ttl
            }
        )

    async def stop_protocol_with_threading(self, thid: str, off_response: bool=False):
        await self.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/stop_protocol',
            params={
                'thid': thid,
                'off_response': off_response
            },
            wait_response=not off_response
        )

    async def stop_protocol_with_threads(self, threads: List[str], off_response: bool=False):
        await self.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/stop_protocol',
            params={
                'threads': threads,
                'off_response': off_response
            },
            wait_response=not off_response
        )

    async def stop_protocol_for_p2p(
            self, sender_verkey: str, recipient_verkey: str, protocols: List[str], off_response: bool=False
    ):
        await self.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/stop_protocol',
            params={
                'sender_verkey': sender_verkey,
                'recipient_verkey': recipient_verkey,
                'protocols': protocols,
                'off_response': off_response
            },
            wait_response=not off_response
        )

    @classmethod
    def _path(cls):
        return '/rpc'

    async def _setup(self, context: Message):
        # Extract proxy info
        proxies = context.get('~proxy', [])
        channel_rpc = None
        channel_sub_protocol = None
        for proxy in proxies:
            if proxy['id'] == 'reverse':
                channel_rpc = proxy['data']['json']['address']
            elif proxy['id'] == 'sub-protocol':
                channel_sub_protocol = proxy['data']['json']['address']
        if channel_rpc is None:
            raise RuntimeError('rpc channel is empty')
        if channel_sub_protocol is None:
            raise RuntimeError('sub-protocol channel is empty')
        self.__tunnel_rpc = AddressedTunnel(
            address=channel_rpc, input_=self._connector, output_=self._connector, p2p=self._p2p
        )
        self.__tunnel_coprotocols = AddressedTunnel(
            address=channel_sub_protocol, input_=self._connector, output_=self._connector, p2p=self._p2p
        )
        # Extract active endpoints
        endpoints = context.get('~endpoints', [])
        endpoint_collection = []
        for endpoint in endpoints:
            body = endpoint['data']['json']
            address = body['address']
            frontend_key = body.get('frontend_routing_key', None)
            if frontend_key:
                for routing_key in body.get('routing_keys', []):
                    is_default = routing_key['is_default']
                    key = routing_key['routing_key']
                    endpoint_collection.append(
                        Endpoint(address=address, routing_keys=[key, frontend_key], is_default=is_default)
                    )
            else:
                endpoint_collection.append(
                    Endpoint(address=address, routing_keys=[], is_default=False)
                )
        if not endpoint_collection:
            raise RuntimeError('Endpoints are empty')
        self.__endpoints = endpoint_collection
        # Extract Networks
        self.__networks = context.get('~networks', [])

    async def _reopen(self):
        await self._connector.reopen()
        payload = await self._connector.read(timeout=1)
        context = Message.deserialize(payload.decode())
        await self._setup(context)

    async def close(self):
        await super().close()
        for ws, session in self.__websockets.values():
            await ws.close()
            await session.close()

    async def __get_websocket(self, url: str):
        tup = self.__websockets.get(url, None)
        if tup is None:
            session = aiohttp.ClientSession(timeout=aiohttp.ClientTimeout(total=self.timeout))
            ws = await session.ws_connect(url=url)
            self.__websockets[url] = (ws, session)
        else:
            ws, session = tup
            if ws.closed:
                ws = session.ws_connect(url=url)
                self.__websockets[url] = (ws, session)
        return ws


class AgentEvents(BaseAgentConnection):
    """RPC service.

    Reactive nature of Smart-Contract design
    """

    RECONNECT_TRY_COUNT = 2

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__tunnel = None
        self.__balancing_group = None

    @property
    def balancing_group(self) -> str:
        return self.__balancing_group

    async def pull(self, timeout: int = None) -> Message:
        if not self._connector.is_open:
            raise SiriusConnectionClosed('Open agent connection at first')
        data = None
        for n in range(self.RECONNECT_TRY_COUNT):
            try:
                data = await self._connector.read(timeout=timeout)
                break
            except SiriusConnectionClosed:
                await self._reopen()
        if data is None:
            SiriusConnectionClosed('agent unreachable')
        try:
            payload = json.loads(data.decode(self._connector.ENC))
        except json.JSONDecodeError:
            raise SiriusInvalidPayloadStructure()
        if 'protected' in payload:
            message = self._p2p.unpack(payload)
            return Message(message)
        else:
            return Message(payload)

    @classmethod
    def _path(cls):
        return '/events'

    async def _reopen(self):
        await self._connector.reopen()
        payload = await self._connector.read(timeout=1)
        context = Message.deserialize(payload.decode())
        await self._setup(context)

    async def _setup(self, context: Message):
        # Extract load balancing info
        balancing = context.get('~balancing', [])
        for balance in balancing:
            group_id = balance.get('data', {}).get('json', {}).get('group_id')
            if group_id:
                self.__balancing_group = group_id



import uuid

import pytest

import sirius_sdk
from sirius_sdk.hub.defaults.default_storage import InMemoryImmutableCollection, InMemoryKeyValueStorage
from sirius_sdk.errors.indy_exceptions import WalletItemNotFound


@pytest.mark.asyncio
async def test_inmemory_kv_storage():
    kv = InMemoryKeyValueStorage()
    await kv.select_db('db1')

    await kv.set('key1', 'value1')
    value = await kv.get('key1')
    assert value == 'value1'

    await kv.select_db('db2')
    await kv.set('key1', 1000)
    value = await kv.get('key1')
    assert value == 1000

    await kv.select_db('db1')
    value = await kv.get('key1')
    assert value == 'value1'

    await kv.delete('key1')
    value = await kv.get('key1')
    assert value is None

    await kv.delete('unknown-key')


@pytest.mark.asyncio
async def test_inmemory_kv_storage_iteration():
    kv = InMemoryKeyValueStorage()
    await kv.select_db('db1')
    data_under_test = {'key1': 'value1', 'key2': 12345}

    for k, v in data_under_test.items():
        await kv.set(k, v)

    loaded_data = await kv.items()
    assert loaded_data == data_under_test


@pytest.mark.asyncio
async def test_inmemory_immutable_collection():
    collection = InMemoryImmutableCollection()

    await collection.select_db('db1')
    await collection.add('Value1', {'tag1': 'tag-val-1', 'tag2': 'tag-val-2'})
    await collection.add('Value2', {'tag1': 'tag-val-1', 'tag2': 'tag-val-3'})

    fetched1 = await collection.fetch({'tag1': 'tag-val-1'})
    assert len(fetched1) == 2

    fetched1 = await collection.fetch({'tag2': 'tag-val-2'})
    assert len(fetched1) == 1
    assert fetched1[0] == 'Value1'

    await collection.select_db('db2')
    fetched3 = await collection.fetch({})
    assert len(fetched3) == 0


@pytest.mark.asyncio
async def test_default_storage_different_space_for_different_hub(config_a: dict, config_b: dict):
    test_key = 'key-' + uuid.uuid4().hex
    test_val = 'test_value'
    typ = 'StorageType'
    async with sirius_sdk.context(**config_a):
        await sirius_sdk.NonSecrets.add_wallet_record(
            type_=typ, id_=test_key, value=test_val
        )
        val = await sirius_sdk.NonSecrets.get_wallet_record(
            type_=typ, id_=test_key, options=sirius_sdk.NonSecretsRetrieveRecordOptions().check_all()
        )
    async with sirius_sdk.context(**config_b):
        with pytest.raises(WalletItemNotFound):
            val = await sirius_sdk.NonSecrets.get_wallet_record(
                type_=typ, id_=test_key, options=sirius_sdk.NonSecretsRetrieveRecordOptions().check_all()
            )



import uuid

import pytest

from sirius_sdk import Agent
from sirius_sdk.agent.storages import InWalletImmutableCollection


@pytest.mark.asyncio
async def test_inwallet_immutable_collection(agent1: Agent):
    await agent1.open()
    try:
        collection = InWalletImmutableCollection(agent1.wallet.non_secrets)

        value1 = {
            'key1': 'value1',
            'key2': 10000
        }
        value2 = {
            'key1': 'value2',
            'key2': 50000
        }

        await collection.select_db(db_name=uuid.uuid4().hex)
        await collection.add(value1, {'tag': 'value1'})
        await collection.add(value2, {'tag': 'value2'})

        fetched, count = await collection.fetch({'tag': 'value1'})
        assert count == 1
        assert len(fetched) == 1
        assert fetched[0] == value1

        fetched, count = await collection.fetch({'tag': 'value2'})
        assert count == 1
        assert len(fetched) == 1
        assert fetched[0] == value2

        fetched, count = await collection.fetch({})
        assert count == 2

        await collection.select_db(db_name=uuid.uuid4().hex)
        fetched, count = await collection.fetch({})
        assert count == 0
    finally:
        await agent1.close()



import uuid

import pytest

from sirius_sdk.errors.exceptions import BaseSiriusException
from sirius_sdk import Agent, Pairwise
from sirius_sdk.agent.aries_rfc.feature_0048_trust_ping import Ping

from tests.helpers import run_coroutines, IndyAgent


@pytest.mark.asyncio
async def test_establish_connection(agent1: Agent, agent2: Agent, agent3: Agent):
    await agent1.open()
    await agent2.open()
    await agent3.open()
    try:
        did1, verkey1 = await agent1.wallet.did.create_and_store_my_did()
        did2, verkey2 = await agent2.wallet.did.create_and_store_my_did()
        endpoint_address_2 = [e for e in agent2.endpoints if e.routing_keys == []][0].address
        endpoint_address_3 = [e for e in agent3.endpoints if e.routing_keys == []][0].address

        await agent1.wallet.did.store_their_did(did2, verkey2)
        await agent1.wallet.pairwise.create_pairwise(did2, did1)
        await agent2.wallet.did.store_their_did(did1, verkey1)
        await agent2.wallet.pairwise.create_pairwise(did1, did2)

        to = Pairwise(
            me=Pairwise.Me(did=did1, verkey=verkey1),
            their=Pairwise.Their(did=did2, label='Agent2', endpoint=endpoint_address_2, verkey=verkey2)
        )
        listener2 = await agent2.subscribe()
        ping = Ping(comment=uuid.uuid4().hex)

        # Check OK
        await agent1.send_to(ping, to)
        event = await listener2.get_one()
        recv = event.message
        assert isinstance(recv, Ping)
        assert recv.comment == ping.comment

    finally:
        await agent1.close()
        await agent2.close()
        await agent3.close()



from sirius_sdk.messaging import Message, Type, validate_common_blocks, register_message_class
from sirius_sdk.errors.exceptions import *

from .mixins import PleaseAckMixin, ThreadMixin, AttachesMixin, TimingMixin
from .decorators import *


VALID_DOC_URI = ['https://didcomm.org/', 'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/']
ARIES_DOC_URI = VALID_DOC_URI[1]


class AriesProtocolMessage(PleaseAckMixin, ThreadMixin, AttachesMixin, TimingMixin, Message):

    DOC_URI = ARIES_DOC_URI
    PROTOCOL = None
    NAME = None
    DEF_VERSION = '1.0'

    def __init__(self, id_: str = None, version: str = None, doc_uri: str = None, *args, **kwargs):
        if self.NAME and ('@type' not in dict(*args, **kwargs)):
            kwargs['@type'] = str(
                Type(
                    doc_uri=doc_uri or self.DOC_URI,
                    protocol=self.PROTOCOL,
                    name=self.NAME, version=version or self.DEF_VERSION
                )
            )
        super().__init__(*args, **kwargs)
        if id_ is not None:
            self['@id'] = id_
        if self.doc_uri not in VALID_DOC_URI:
            raise SiriusValidationError('Unexpected doc_uri "%s"' % self.doc_uri)
        if self.protocol != self.PROTOCOL:
            raise SiriusValidationError('Unexpected protocol "%s"' % self.protocol)
        if self.name != self.NAME:
            raise SiriusValidationError('Unexpected name "%s"' % self.name)

    def validate(self):
        validate_common_blocks(self)


class AriesProblemReport(AriesProtocolMessage):

    NAME = 'problem_report'

    def __init__(self, problem_code: str=None, explain: str=None, thread_id: str=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if problem_code:
            self['problem-code'] = problem_code
        if explain:
            self['explain'] = explain
        if thread_id is not None:
            thread = self.get(THREAD_DECORATOR, {})
            thread['thid'] = thread_id
            self[THREAD_DECORATOR] = thread

    @property
    def problem_code(self) -> str:
        return self.get('problem-code', '')

    @property
    def explain(self) -> str:
        return self.get('explain', '')


class RegisterMessage(type):

    def __new__(meta, name, bases, class_dict):
        cls = type.__new__(meta, name, bases, class_dict)
        if issubclass(cls, AriesProtocolMessage):
            register_message_class(cls, protocol=cls.PROTOCOL, name=cls.NAME)
        return cls





r"""DIDComm Transports are simplex: they only transfer messages from sender to receiver,
using the endpoint present in the receiver's DID Document.
This extension enables bi-directional communication on the same transports, even when one party has no public endpoint.
Messages can flow back in response to inbound messages over the same connection.
This is particularly useful for communication between agents that are unable to provide routable endpoints
(such as mobile phones or agents inside a firewall) and their mediators.

Details: https://github.com/decentralized-identity/didcomm-messaging/blob/master/extensions/return_route/main.md
"""
from .headers import HEADER_NAME, extract_value, RouteType
from .const import URI_QUEUE_TRANSPORT


__all__ = ["extract_value", "HEADER_NAME", "RouteType", "URI_QUEUE_TRANSPORT"]
